# -*- coding: utf-8 -*-
# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from __future__ import annotations

from typing import MutableMapping, MutableSequence

from google.protobuf import duration_pb2  # type: ignore
from google.protobuf import wrappers_pb2  # type: ignore
from google.rpc import status_pb2  # type: ignore
from google.type import timeofday_pb2  # type: ignore
import proto  # type: ignore

from google.cloud.monitoring_v3.types import mutation_record as gm_mutation_record
from google.cloud.monitoring_v3.types import common

__protobuf__ = proto.module(
    package="google.monitoring.v3",
    manifest={
        "AlertPolicy",
    },
)


class AlertPolicy(proto.Message):
    r"""A description of the conditions under which some aspect of your
    system is considered to be "unhealthy" and the ways to notify people
    or services about this state. For an overview of alerting policies,
    see `Introduction to
    Alerting <https://cloud.google.com/monitoring/alerts/>`__.

    Attributes:
        name (str):
            Identifier. Required if the policy exists. The resource name
            for this policy. The format is:

            ::

                projects/[PROJECT_ID_OR_NUMBER]/alertPolicies/[ALERT_POLICY_ID]

            ``[ALERT_POLICY_ID]`` is assigned by Cloud Monitoring when
            the policy is created. When calling the
            [alertPolicies.create][google.monitoring.v3.AlertPolicyService.CreateAlertPolicy]
            method, do not include the ``name`` field in the alerting
            policy passed as part of the request.
        display_name (str):
            A short name or phrase used to identify the policy in
            dashboards, notifications, and incidents. To avoid
            confusion, don't use the same display name for multiple
            policies in the same project. The name is limited to 512
            Unicode characters.

            The convention for the display_name of a
            PrometheusQueryLanguageCondition is "{rule group
            name}/{alert name}", where the {rule group name} and {alert
            name} should be taken from the corresponding Prometheus
            configuration file. This convention is not enforced. In any
            case the display_name is not a unique key of the
            AlertPolicy.
        documentation (google.cloud.monitoring_v3.types.AlertPolicy.Documentation):
            Documentation that is included with
            notifications and incidents related to this
            policy. Best practice is for the documentation
            to include information to help responders
            understand, mitigate, escalate, and correct the
            underlying problems detected by the alerting
            policy. Notification channels that have limited
            capacity might not show this documentation.
        user_labels (MutableMapping[str, str]):
            User-supplied key/value data to be used for organizing and
            identifying the ``AlertPolicy`` objects.

            The field can contain up to 64 entries. Each key and value
            is limited to 63 Unicode characters or 128 bytes, whichever
            is smaller. Labels and values can contain only lowercase
            letters, numerals, underscores, and dashes. Keys must begin
            with a letter.

            Note that Prometheus {alert name} is a `valid Prometheus
            label
            names <https://prometheus.io/docs/concepts/data_model/#metric-names-and-labels>`__,
            whereas Prometheus {rule group} is an unrestricted UTF-8
            string. This means that they cannot be stored as-is in user
            labels, because they may contain characters that are not
            allowed in user-label values.
        conditions (MutableSequence[google.cloud.monitoring_v3.types.AlertPolicy.Condition]):
            A list of conditions for the policy. The conditions are
            combined by AND or OR according to the ``combiner`` field.
            If the combined conditions evaluate to true, then an
            incident is created. A policy can have from one to six
            conditions. If ``condition_time_series_query_language`` is
            present, it must be the only ``condition``. If
            ``condition_monitoring_query_language`` is present, it must
            be the only ``condition``.
        combiner (google.cloud.monitoring_v3.types.AlertPolicy.ConditionCombinerType):
            How to combine the results of multiple conditions to
            determine if an incident should be opened. If
            ``condition_time_series_query_language`` is present, this
            must be ``COMBINE_UNSPECIFIED``.
        enabled (google.protobuf.wrappers_pb2.BoolValue):
            Whether or not the policy is enabled. On
            write, the default interpretation if unset is
            that the policy is enabled. On read, clients
            should not make any assumption about the state
            if it has not been populated. The field should
            always be populated on List and Get operations,
            unless a field projection has been specified
            that strips it out.
        validity (google.rpc.status_pb2.Status):
            Read-only description of how the alerting
            policy is invalid. This field is only set when
            the alerting policy is invalid. An invalid
            alerting policy will not generate incidents.
        notification_channels (MutableSequence[str]):
            Identifies the notification channels to which notifications
            should be sent when incidents are opened or closed or when
            new violations occur on an already opened incident. Each
            element of this array corresponds to the ``name`` field in
            each of the
            [``NotificationChannel``][google.monitoring.v3.NotificationChannel]
            objects that are returned from the
            [``ListNotificationChannels``]
            [google.monitoring.v3.NotificationChannelService.ListNotificationChannels]
            method. The format of the entries in this field is:

            ::

                projects/[PROJECT_ID_OR_NUMBER]/notificationChannels/[CHANNEL_ID]
        creation_record (google.cloud.monitoring_v3.types.MutationRecord):
            A read-only record of the creation of the
            alerting policy. If provided in a call to create
            or update, this field will be ignored.
        mutation_record (google.cloud.monitoring_v3.types.MutationRecord):
            A read-only record of the most recent change
            to the alerting policy. If provided in a call to
            create or update, this field will be ignored.
        alert_strategy (google.cloud.monitoring_v3.types.AlertPolicy.AlertStrategy):
            Control over how this alerting policy's
            notification channels are notified.
        severity (google.cloud.monitoring_v3.types.AlertPolicy.Severity):
            Optional. The severity of an alerting policy
            indicates how important incidents generated by
            that policy are. The severity level will be
            displayed on the Incident detail page and in
            notifications.
    """

    class ConditionCombinerType(proto.Enum):
        r"""Operators for combining conditions.

        Values:
            COMBINE_UNSPECIFIED (0):
                An unspecified combiner.
            AND (1):
                Combine conditions using the logical ``AND`` operator. An
                incident is created only if all the conditions are met
                simultaneously. This combiner is satisfied if all conditions
                are met, even if they are met on completely different
                resources.
            OR (2):
                Combine conditions using the logical ``OR`` operator. An
                incident is created if any of the listed conditions is met.
            AND_WITH_MATCHING_RESOURCE (3):
                Combine conditions using logical ``AND`` operator, but
                unlike the regular ``AND`` option, an incident is created
                only if all conditions are met simultaneously on at least
                one resource.
        """
        COMBINE_UNSPECIFIED = 0
        AND = 1
        OR = 2
        AND_WITH_MATCHING_RESOURCE = 3

    class Severity(proto.Enum):
        r"""An enumeration of possible severity level for an alerting
        policy.

        Values:
            SEVERITY_UNSPECIFIED (0):
                No severity is specified. This is the default
                value.
            CRITICAL (1):
                This is the highest severity level. Use this
                if the problem could cause significant damage or
                downtime.
            ERROR (2):
                This is the medium severity level. Use this
                if the problem could cause minor damage or
                downtime.
            WARNING (3):
                This is the lowest severity level. Use this
                if the problem is not causing any damage or
                downtime, but could potentially lead to a
                problem in the future.
        """
        SEVERITY_UNSPECIFIED = 0
        CRITICAL = 1
        ERROR = 2
        WARNING = 3

    class Documentation(proto.Message):
        r"""Documentation that is included in the notifications and
        incidents pertaining to this policy.

        Attributes:
            content (str):
                The body of the documentation, interpreted according to
                ``mime_type``. The content may not exceed 8,192 Unicode
                characters and may not exceed more than 10,240 bytes when
                encoded in UTF-8 format, whichever is smaller. This text can
                be `templatized by using
                variables <https://cloud.google.com/monitoring/alerts/doc-variables#doc-vars>`__.
            mime_type (str):
                The format of the ``content`` field. Presently, only the
                value ``"text/markdown"`` is supported. See
                `Markdown <https://en.wikipedia.org/wiki/Markdown>`__ for
                more information.
            subject (str):
                Optional. The subject line of the notification. The subject
                line may not exceed 10,240 bytes. In notifications generated
                by this policy, the contents of the subject line after
                variable expansion will be truncated to 255 bytes or shorter
                at the latest UTF-8 character boundary. The 255-byte limit
                is recommended by `this
                thread <https://stackoverflow.com/questions/1592291/what-is-the-email-subject-length-limit>`__.
                It is both the limit imposed by some third-party ticketing
                products and it is common to define textual fields in
                databases as VARCHAR(255).

                The contents of the subject line can be `templatized by
                using
                variables <https://cloud.google.com/monitoring/alerts/doc-variables#doc-vars>`__.
                If this field is missing or empty, a default subject line
                will be generated.
            links (MutableSequence[google.cloud.monitoring_v3.types.AlertPolicy.Documentation.Link]):
                Optional. Links to content such as playbooks,
                repositories, and other resources. This field
                can contain up to 3 entries.
        """

        class Link(proto.Message):
            r"""Links to content such as playbooks, repositories, and other
            resources.

            Attributes:
                display_name (str):
                    A short display name for the link. The
                    display name must not be empty or exceed 63
                    characters. Example: "playbook".
                url (str):
                    The url of a webpage. A url can be templatized by using
                    variables in the path or the query parameters. The total
                    length of a URL should not exceed 2083 characters before and
                    after variable expansion. Example:
                    "https://my_domain.com/playbook?name=${resource.name}".
            """

            display_name: str = proto.Field(
                proto.STRING,
                number=1,
            )
            url: str = proto.Field(
                proto.STRING,
                number=2,
            )

        content: str = proto.Field(
            proto.STRING,
            number=1,
        )
        mime_type: str = proto.Field(
            proto.STRING,
            number=2,
        )
        subject: str = proto.Field(
            proto.STRING,
            number=3,
        )
        links: MutableSequence["AlertPolicy.Documentation.Link"] = proto.RepeatedField(
            proto.MESSAGE,
            number=4,
            message="AlertPolicy.Documentation.Link",
        )

    class Condition(proto.Message):
        r"""A condition is a true/false test that determines when an
        alerting policy should open an incident. If a condition
        evaluates to true, it signifies that something is wrong.

        This message has `oneof`_ fields (mutually exclusive fields).
        For each oneof, at most one member field can be set at the same time.
        Setting any member of the oneof automatically clears all other
        members.

        .. _oneof: https://proto-plus-python.readthedocs.io/en/stable/fields.html#oneofs-mutually-exclusive-fields

        Attributes:
            name (str):
                Required if the condition exists. The unique resource name
                for this condition. Its format is:

                ::

                    projects/[PROJECT_ID_OR_NUMBER]/alertPolicies/[POLICY_ID]/conditions/[CONDITION_ID]

                ``[CONDITION_ID]`` is assigned by Cloud Monitoring when the
                condition is created as part of a new or updated alerting
                policy.

                When calling the
                [alertPolicies.create][google.monitoring.v3.AlertPolicyService.CreateAlertPolicy]
                method, do not include the ``name`` field in the conditions
                of the requested alerting policy. Cloud Monitoring creates
                the condition identifiers and includes them in the new
                policy.

                When calling the
                [alertPolicies.update][google.monitoring.v3.AlertPolicyService.UpdateAlertPolicy]
                method to update a policy, including a condition ``name``
                causes the existing condition to be updated. Conditions
                without names are added to the updated policy. Existing
                conditions are deleted if they are not updated.

                Best practice is to preserve ``[CONDITION_ID]`` if you make
                only small changes, such as those to condition thresholds,
                durations, or trigger values. Otherwise, treat the change as
                a new condition and let the existing condition be deleted.
            display_name (str):
                A short name or phrase used to identify the
                condition in dashboards, notifications, and
                incidents. To avoid confusion, don't use the
                same display name for multiple conditions in the
                same policy.
            condition_threshold (google.cloud.monitoring_v3.types.AlertPolicy.Condition.MetricThreshold):
                A condition that compares a time series
                against a threshold.

                This field is a member of `oneof`_ ``condition``.
            condition_absent (google.cloud.monitoring_v3.types.AlertPolicy.Condition.MetricAbsence):
                A condition that checks that a time series
                continues to receive new data points.

                This field is a member of `oneof`_ ``condition``.
            condition_matched_log (google.cloud.monitoring_v3.types.AlertPolicy.Condition.LogMatch):
                A condition that checks for log messages
                matching given constraints. If set, no other
                conditions can be present.

                This field is a member of `oneof`_ ``condition``.
            condition_monitoring_query_language (google.cloud.monitoring_v3.types.AlertPolicy.Condition.MonitoringQueryLanguageCondition):
                A condition that uses the Monitoring Query
                Language to define alerts.

                This field is a member of `oneof`_ ``condition``.
            condition_prometheus_query_language (google.cloud.monitoring_v3.types.AlertPolicy.Condition.PrometheusQueryLanguageCondition):
                A condition that uses the Prometheus query
                language to define alerts.

                This field is a member of `oneof`_ ``condition``.
            condition_sql (google.cloud.monitoring_v3.types.AlertPolicy.Condition.SqlCondition):
                A condition that periodically evaluates a SQL
                query result.

                This field is a member of `oneof`_ ``condition``.
        """

        class EvaluationMissingData(proto.Enum):
            r"""A condition control that determines how metric-threshold
            conditions are evaluated when data stops arriving.
            This control doesn't affect metric-absence policies.

            Values:
                EVALUATION_MISSING_DATA_UNSPECIFIED (0):
                    An unspecified evaluation missing data option. Equivalent to
                    EVALUATION_MISSING_DATA_NO_OP.
                EVALUATION_MISSING_DATA_INACTIVE (1):
                    If there is no data to evaluate the
                    condition, then evaluate the condition as false.
                EVALUATION_MISSING_DATA_ACTIVE (2):
                    If there is no data to evaluate the
                    condition, then evaluate the condition as true.
                EVALUATION_MISSING_DATA_NO_OP (3):
                    Do not evaluate the condition to any value if
                    there is no data.
            """
            EVALUATION_MISSING_DATA_UNSPECIFIED = 0
            EVALUATION_MISSING_DATA_INACTIVE = 1
            EVALUATION_MISSING_DATA_ACTIVE = 2
            EVALUATION_MISSING_DATA_NO_OP = 3

        class Trigger(proto.Message):
            r"""Specifies how many time series must fail a predicate to trigger a
            condition. If not specified, then a ``{count: 1}`` trigger is used.

            This message has `oneof`_ fields (mutually exclusive fields).
            For each oneof, at most one member field can be set at the same time.
            Setting any member of the oneof automatically clears all other
            members.

            .. _oneof: https://proto-plus-python.readthedocs.io/en/stable/fields.html#oneofs-mutually-exclusive-fields

            Attributes:
                count (int):
                    The absolute number of time series that must
                    fail the predicate for the condition to be
                    triggered.

                    This field is a member of `oneof`_ ``type``.
                percent (float):
                    The percentage of time series that must fail
                    the predicate for the condition to be triggered.

                    This field is a member of `oneof`_ ``type``.
            """

            count: int = proto.Field(
                proto.INT32,
                number=1,
                oneof="type",
            )
            percent: float = proto.Field(
                proto.DOUBLE,
                number=2,
                oneof="type",
            )

        class MetricThreshold(proto.Message):
            r"""A condition type that compares a collection of time series
            against a threshold.

            Attributes:
                filter (str):
                    Required. A
                    `filter <https://cloud.google.com/monitoring/api/v3/filters>`__
                    that identifies which time series should be compared with
                    the threshold.

                    The filter is similar to the one that is specified in the
                    ```ListTimeSeries``
                    request <https://cloud.google.com/monitoring/api/ref_v3/rest/v3/projects.timeSeries/list>`__
                    (that call is useful to verify the time series that will be
                    retrieved / processed). The filter must specify the metric
                    type and the resource type. Optionally, it can specify
                    resource labels and metric labels. This field must not
                    exceed 2048 Unicode characters in length.
                aggregations (MutableSequence[google.cloud.monitoring_v3.types.Aggregation]):
                    Specifies the alignment of data points in individual time
                    series as well as how to combine the retrieved time series
                    together (such as when aggregating multiple streams on each
                    resource to a single stream for each resource or when
                    aggregating streams across all members of a group of
                    resources). Multiple aggregations are applied in the order
                    specified.

                    This field is similar to the one in the ```ListTimeSeries``
                    request <https://cloud.google.com/monitoring/api/ref_v3/rest/v3/projects.timeSeries/list>`__.
                    It is advisable to use the ``ListTimeSeries`` method when
                    debugging this field.
                denominator_filter (str):
                    A
                    `filter <https://cloud.google.com/monitoring/api/v3/filters>`__
                    that identifies a time series that should be used as the
                    denominator of a ratio that will be compared with the
                    threshold. If a ``denominator_filter`` is specified, the
                    time series specified by the ``filter`` field will be used
                    as the numerator.

                    The filter must specify the metric type and optionally may
                    contain restrictions on resource type, resource labels, and
                    metric labels. This field may not exceed 2048 Unicode
                    characters in length.
                denominator_aggregations (MutableSequence[google.cloud.monitoring_v3.types.Aggregation]):
                    Specifies the alignment of data points in individual time
                    series selected by ``denominatorFilter`` as well as how to
                    combine the retrieved time series together (such as when
                    aggregating multiple streams on each resource to a single
                    stream for each resource or when aggregating streams across
                    all members of a group of resources).

                    When computing ratios, the ``aggregations`` and
                    ``denominator_aggregations`` fields must use the same
                    alignment period and produce time series that have the same
                    periodicity and labels.
                forecast_options (google.cloud.monitoring_v3.types.AlertPolicy.Condition.MetricThreshold.ForecastOptions):
                    When this field is present, the ``MetricThreshold``
                    condition forecasts whether the time series is predicted to
                    violate the threshold within the ``forecast_horizon``. When
                    this field is not set, the ``MetricThreshold`` tests the
                    current value of the timeseries against the threshold.
                comparison (google.cloud.monitoring_v3.types.ComparisonType):
                    The comparison to apply between the time series (indicated
                    by ``filter`` and ``aggregation``) and the threshold
                    (indicated by ``threshold_value``). The comparison is
                    applied on each time series, with the time series on the
                    left-hand side and the threshold on the right-hand side.

                    Only ``COMPARISON_LT`` and ``COMPARISON_GT`` are supported
                    currently.
                threshold_value (float):
                    A value against which to compare the time
                    series.
                duration (google.protobuf.duration_pb2.Duration):
                    The amount of time that a time series must violate the
                    threshold to be considered failing. Currently, only values
                    that are a multiple of a minute--e.g., 0, 60, 120, or 300
                    seconds--are supported. If an invalid value is given, an
                    error will be returned. When choosing a duration, it is
                    useful to keep in mind the frequency of the underlying time
                    series data (which may also be affected by any alignments
                    specified in the ``aggregations`` field); a good duration is
                    long enough so that a single outlier does not generate
                    spurious alerts, but short enough that unhealthy states are
                    detected and alerted on quickly.
                trigger (google.cloud.monitoring_v3.types.AlertPolicy.Condition.Trigger):
                    The number/percent of time series for which the comparison
                    must hold in order for the condition to trigger. If
                    unspecified, then the condition will trigger if the
                    comparison is true for any of the time series that have been
                    identified by ``filter`` and ``aggregations``, or by the
                    ratio, if ``denominator_filter`` and
                    ``denominator_aggregations`` are specified.
                evaluation_missing_data (google.cloud.monitoring_v3.types.AlertPolicy.Condition.EvaluationMissingData):
                    A condition control that determines how metric-threshold
                    conditions are evaluated when data stops arriving. To use
                    this control, the value of the ``duration`` field must be
                    greater than or equal to 60 seconds.
            """

            class ForecastOptions(proto.Message):
                r"""Options used when forecasting the time series and testing
                the predicted value against the threshold.

                Attributes:
                    forecast_horizon (google.protobuf.duration_pb2.Duration):
                        Required. The length of time into the future to forecast
                        whether a time series will violate the threshold. If the
                        predicted value is found to violate the threshold, and the
                        violation is observed in all forecasts made for the
                        configured ``duration``, then the time series is considered
                        to be failing. The forecast horizon can range from 1 hour to
                        60 hours.
                """

                forecast_horizon: duration_pb2.Duration = proto.Field(
                    proto.MESSAGE,
                    number=1,
                    message=duration_pb2.Duration,
                )

            filter: str = proto.Field(
                proto.STRING,
                number=2,
            )
            aggregations: MutableSequence[common.Aggregation] = proto.RepeatedField(
                proto.MESSAGE,
                number=8,
                message=common.Aggregation,
            )
            denominator_filter: str = proto.Field(
                proto.STRING,
                number=9,
            )
            denominator_aggregations: MutableSequence[
                common.Aggregation
            ] = proto.RepeatedField(
                proto.MESSAGE,
                number=10,
                message=common.Aggregation,
            )
            forecast_options: "AlertPolicy.Condition.MetricThreshold.ForecastOptions" = proto.Field(
                proto.MESSAGE,
                number=12,
                message="AlertPolicy.Condition.MetricThreshold.ForecastOptions",
            )
            comparison: common.ComparisonType = proto.Field(
                proto.ENUM,
                number=4,
                enum=common.ComparisonType,
            )
            threshold_value: float = proto.Field(
                proto.DOUBLE,
                number=5,
            )
            duration: duration_pb2.Duration = proto.Field(
                proto.MESSAGE,
                number=6,
                message=duration_pb2.Duration,
            )
            trigger: "AlertPolicy.Condition.Trigger" = proto.Field(
                proto.MESSAGE,
                number=7,
                message="AlertPolicy.Condition.Trigger",
            )
            evaluation_missing_data: "AlertPolicy.Condition.EvaluationMissingData" = (
                proto.Field(
                    proto.ENUM,
                    number=11,
                    enum="AlertPolicy.Condition.EvaluationMissingData",
                )
            )

        class MetricAbsence(proto.Message):
            r"""A condition type that checks that monitored resources are reporting
            data. The configuration defines a metric and a set of monitored
            resources. The predicate is considered in violation when a time
            series for the specified metric of a monitored resource does not
            include any data in the specified ``duration``.

            Attributes:
                filter (str):
                    Required. A
                    `filter <https://cloud.google.com/monitoring/api/v3/filters>`__
                    that identifies which time series should be compared with
                    the threshold.

                    The filter is similar to the one that is specified in the
                    ```ListTimeSeries``
                    request <https://cloud.google.com/monitoring/api/ref_v3/rest/v3/projects.timeSeries/list>`__
                    (that call is useful to verify the time series that will be
                    retrieved / processed). The filter must specify the metric
                    type and the resource type. Optionally, it can specify
                    resource labels and metric labels. This field must not
                    exceed 2048 Unicode characters in length.
                aggregations (MutableSequence[google.cloud.monitoring_v3.types.Aggregation]):
                    Specifies the alignment of data points in individual time
                    series as well as how to combine the retrieved time series
                    together (such as when aggregating multiple streams on each
                    resource to a single stream for each resource or when
                    aggregating streams across all members of a group of
                    resources). Multiple aggregations are applied in the order
                    specified.

                    This field is similar to the one in the ```ListTimeSeries``
                    request <https://cloud.google.com/monitoring/api/ref_v3/rest/v3/projects.timeSeries/list>`__.
                    It is advisable to use the ``ListTimeSeries`` method when
                    debugging this field.
                duration (google.protobuf.duration_pb2.Duration):
                    The amount of time that a time series must fail to report
                    new data to be considered failing. The minimum value of this
                    field is 120 seconds. Larger values that are a multiple of a
                    minute--for example, 240 or 300 seconds--are supported. If
                    an invalid value is given, an error will be returned. The
                    ``Duration.nanos`` field is ignored.
                trigger (google.cloud.monitoring_v3.types.AlertPolicy.Condition.Trigger):
                    The number/percent of time series for which the comparison
                    must hold in order for the condition to trigger. If
                    unspecified, then the condition will trigger if the
                    comparison is true for any of the time series that have been
                    identified by ``filter`` and ``aggregations``.
            """

            filter: str = proto.Field(
                proto.STRING,
                number=1,
            )
            aggregations: MutableSequence[common.Aggregation] = proto.RepeatedField(
                proto.MESSAGE,
                number=5,
                message=common.Aggregation,
            )
            duration: duration_pb2.Duration = proto.Field(
                proto.MESSAGE,
                number=2,
                message=duration_pb2.Duration,
            )
            trigger: "AlertPolicy.Condition.Trigger" = proto.Field(
                proto.MESSAGE,
                number=3,
                message="AlertPolicy.Condition.Trigger",
            )

        class LogMatch(proto.Message):
            r"""A condition type that checks whether a log message in the `scoping
            project <https://cloud.google.com/monitoring/api/v3#project_name>`__
            satisfies the given filter. Logs from other projects in the metrics
            scope are not evaluated.

            Attributes:
                filter (str):
                    Required. A logs-based filter. See `Advanced Logs
                    Queries <https://cloud.google.com/logging/docs/view/advanced-queries>`__
                    for how this filter should be constructed.
                label_extractors (MutableMapping[str, str]):
                    Optional. A map from a label key to an extractor expression,
                    which is used to extract the value for this label key. Each
                    entry in this map is a specification for how data should be
                    extracted from log entries that match ``filter``. Each
                    combination of extracted values is treated as a separate
                    rule for the purposes of triggering notifications. Label
                    keys and corresponding values can be used in notifications
                    generated by this condition.

                    Please see `the documentation on logs-based metric
                    ``valueExtractor``\ s <https://cloud.google.com/logging/docs/reference/v2/rest/v2/projects.metrics#LogMetric.FIELDS.value_extractor>`__
                    for syntax and examples.
            """

            filter: str = proto.Field(
                proto.STRING,
                number=1,
            )
            label_extractors: MutableMapping[str, str] = proto.MapField(
                proto.STRING,
                proto.STRING,
                number=2,
            )

        class MonitoringQueryLanguageCondition(proto.Message):
            r"""A condition type that allows alerting policies to be defined using
            `Monitoring Query
            Language <https://cloud.google.com/monitoring/mql>`__.

            Attributes:
                query (str):
                    `Monitoring Query
                    Language <https://cloud.google.com/monitoring/mql>`__ query
                    that outputs a boolean stream.
                duration (google.protobuf.duration_pb2.Duration):
                    The amount of time that a time series must violate the
                    threshold to be considered failing. Currently, only values
                    that are a multiple of a minute--e.g., 0, 60, 120, or 300
                    seconds--are supported. If an invalid value is given, an
                    error will be returned. When choosing a duration, it is
                    useful to keep in mind the frequency of the underlying time
                    series data (which may also be affected by any alignments
                    specified in the ``aggregations`` field); a good duration is
                    long enough so that a single outlier does not generate
                    spurious alerts, but short enough that unhealthy states are
                    detected and alerted on quickly.
                trigger (google.cloud.monitoring_v3.types.AlertPolicy.Condition.Trigger):
                    The number/percent of time series for which the comparison
                    must hold in order for the condition to trigger. If
                    unspecified, then the condition will trigger if the
                    comparison is true for any of the time series that have been
                    identified by ``filter`` and ``aggregations``, or by the
                    ratio, if ``denominator_filter`` and
                    ``denominator_aggregations`` are specified.
                evaluation_missing_data (google.cloud.monitoring_v3.types.AlertPolicy.Condition.EvaluationMissingData):
                    A condition control that determines how
                    metric-threshold conditions are evaluated when
                    data stops arriving.
            """

            query: str = proto.Field(
                proto.STRING,
                number=1,
            )
            duration: duration_pb2.Duration = proto.Field(
                proto.MESSAGE,
                number=2,
                message=duration_pb2.Duration,
            )
            trigger: "AlertPolicy.Condition.Trigger" = proto.Field(
                proto.MESSAGE,
                number=3,
                message="AlertPolicy.Condition.Trigger",
            )
            evaluation_missing_data: "AlertPolicy.Condition.EvaluationMissingData" = (
                proto.Field(
                    proto.ENUM,
                    number=4,
                    enum="AlertPolicy.Condition.EvaluationMissingData",
                )
            )

        class PrometheusQueryLanguageCondition(proto.Message):
            r"""A condition type that allows alerting policies to be defined using
            `Prometheus Query Language
            (PromQL) <https://prometheus.io/docs/prometheus/latest/querying/basics/>`__.

            The PrometheusQueryLanguageCondition message contains information
            from a Prometheus alerting rule and its associated rule group.

            A Prometheus alerting rule is described
            `here <https://prometheus.io/docs/prometheus/latest/configuration/alerting_rules/>`__.
            The semantics of a Prometheus alerting rule is described
            `here <https://prometheus.io/docs/prometheus/latest/configuration/recording_rules/#rule>`__.

            A Prometheus rule group is described
            `here <https://prometheus.io/docs/prometheus/latest/configuration/recording_rules/>`__.
            The semantics of a Prometheus rule group is described
            `here <https://prometheus.io/docs/prometheus/latest/configuration/recording_rules/#rule_group>`__.

            Because Cloud Alerting has no representation of a Prometheus rule
            group resource, we must embed the information of the parent rule
            group inside each of the conditions that refer to it. We must also
            update the contents of all Prometheus alerts in case the information
            of their rule group changes.

            The PrometheusQueryLanguageCondition protocol buffer combines the
            information of the corresponding rule group and alerting rule. The
            structure of the PrometheusQueryLanguageCondition protocol buffer
            does NOT mimic the structure of the Prometheus rule group and
            alerting rule YAML declarations. The
            PrometheusQueryLanguageCondition protocol buffer may change in the
            future to support future rule group and/or alerting rule features.
            There are no new such features at the present time (2023-06-26).

            Attributes:
                query (str):
                    Required. The PromQL expression to evaluate.
                    Every evaluation cycle this expression is
                    evaluated at the current time, and all resultant
                    time series become pending/firing alerts. This
                    field must not be empty.
                duration (google.protobuf.duration_pb2.Duration):
                    Optional. Alerts are considered firing once
                    their PromQL expression was evaluated to be
                    "true" for this long. Alerts whose PromQL
                    expression was not evaluated to be "true" for
                    long enough are considered pending.
                    Must be a non-negative duration or missing.
                    This field is optional. Its default value is
                    zero.
                evaluation_interval (google.protobuf.duration_pb2.Duration):
                    Optional. How often this rule should be
                    evaluated. Must be a positive multiple of 30
                    seconds or missing. This field is optional. Its
                    default value is 30 seconds. If this
                    PrometheusQueryLanguageCondition was generated
                    from a Prometheus alerting rule, then this value
                    should be taken from the enclosing rule group.
                labels (MutableMapping[str, str]):
                    Optional. Labels to add to or overwrite in the PromQL query
                    result. Label names `must be
                    valid <https://prometheus.io/docs/concepts/data_model/#metric-names-and-labels>`__.
                    Label values can be `templatized by using
                    variables <https://cloud.google.com/monitoring/alerts/doc-variables#doc-vars>`__.
                    The only available variable names are the names of the
                    labels in the PromQL result, including "**name**" and
                    "value". "labels" may be empty.
                rule_group (str):
                    Optional. The rule group name of this alert
                    in the corresponding Prometheus configuration
                    file.

                    Some external tools may require this field to be
                    populated correctly in order to refer to the
                    original Prometheus configuration file. The rule
                    group name and the alert name are necessary to
                    update the relevant AlertPolicies in case the
                    definition of the rule group changes in the
                    future.

                    This field is optional. If this field is not
                    empty, then it must contain a valid UTF-8
                    string.
                    This field may not exceed 2048 Unicode
                    characters in length.
                alert_rule (str):
                    Optional. The alerting rule name of this alert in the
                    corresponding Prometheus configuration file.

                    Some external tools may require this field to be populated
                    correctly in order to refer to the original Prometheus
                    configuration file. The rule group name and the alert name
                    are necessary to update the relevant AlertPolicies in case
                    the definition of the rule group changes in the future.

                    This field is optional. If this field is not empty, then it
                    must be a `valid Prometheus label
                    name <https://prometheus.io/docs/concepts/data_model/#metric-names-and-labels>`__.
                    This field may not exceed 2048 Unicode characters in length.
                disable_metric_validation (bool):
                    Optional. Whether to disable metric existence validation for
                    this condition.

                    This allows alerting policies to be defined on metrics that
                    do not yet exist, improving advanced customer workflows such
                    as configuring alerting policies using Terraform.

                    Users with the ``monitoring.alertPolicyViewer`` role are
                    able to see the name of the non-existent metric in the
                    alerting policy condition.
            """

            query: str = proto.Field(
                proto.STRING,
                number=1,
            )
            duration: duration_pb2.Duration = proto.Field(
                proto.MESSAGE,
                number=2,
                message=duration_pb2.Duration,
            )
            evaluation_interval: duration_pb2.Duration = proto.Field(
                proto.MESSAGE,
                number=3,
                message=duration_pb2.Duration,
            )
            labels: MutableMapping[str, str] = proto.MapField(
                proto.STRING,
                proto.STRING,
                number=4,
            )
            rule_group: str = proto.Field(
                proto.STRING,
                number=5,
            )
            alert_rule: str = proto.Field(
                proto.STRING,
                number=6,
            )
            disable_metric_validation: bool = proto.Field(
                proto.BOOL,
                number=7,
            )

        class SqlCondition(proto.Message):
            r"""A condition that allows alerting policies to be defined using
            GoogleSQL. SQL conditions examine a sliding window of logs using
            GoogleSQL. Alert policies with SQL conditions may incur
            additional billing.

            This message has `oneof`_ fields (mutually exclusive fields).
            For each oneof, at most one member field can be set at the same time.
            Setting any member of the oneof automatically clears all other
            members.

            .. _oneof: https://proto-plus-python.readthedocs.io/en/stable/fields.html#oneofs-mutually-exclusive-fields

            Attributes:
                query (str):
                    Required. The Log Analytics SQL query to run, as a string.
                    The query must conform to the required shape. Specifically,
                    the query must not try to filter the input by time. A filter
                    will automatically be applied to filter the input so that
                    the query receives all rows received since the last time the
                    query was run.

                    For example, the following query extracts all log entries
                    containing an HTTP request:

                    ::

                        SELECT
                          timestamp, log_name, severity, http_request, resource, labels
                        FROM
                          my-project.global._Default._AllLogs
                        WHERE
                          http_request IS NOT NULL
                minutes (google.cloud.monitoring_v3.types.AlertPolicy.Condition.SqlCondition.Minutes):
                    Schedule the query to execute every so many
                    minutes.

                    This field is a member of `oneof`_ ``schedule``.
                hourly (google.cloud.monitoring_v3.types.AlertPolicy.Condition.SqlCondition.Hourly):
                    Schedule the query to execute every so many
                    hours.

                    This field is a member of `oneof`_ ``schedule``.
                daily (google.cloud.monitoring_v3.types.AlertPolicy.Condition.SqlCondition.Daily):
                    Schedule the query to execute every so many
                    days.

                    This field is a member of `oneof`_ ``schedule``.
                row_count_test (google.cloud.monitoring_v3.types.AlertPolicy.Condition.SqlCondition.RowCountTest):
                    Test the row count against a threshold.

                    This field is a member of `oneof`_ ``evaluate``.
                boolean_test (google.cloud.monitoring_v3.types.AlertPolicy.Condition.SqlCondition.BooleanTest):
                    Test the boolean value in the indicated
                    column.

                    This field is a member of `oneof`_ ``evaluate``.
            """

            class Minutes(proto.Message):
                r"""Used to schedule the query to run every so many minutes.

                Attributes:
                    periodicity (int):
                        Required. Number of minutes between runs. The
                        interval must be greater than or equal to 5
                        minutes and less than or equal to 1440 minutes.
                """

                periodicity: int = proto.Field(
                    proto.INT32,
                    number=1,
                )

            class Hourly(proto.Message):
                r"""Used to schedule the query to run every so many hours.

                .. _oneof: https://proto-plus-python.readthedocs.io/en/stable/fields.html#oneofs-mutually-exclusive-fields

                Attributes:
                    periodicity (int):
                        Required. The number of hours between runs.
                        Must be greater than or equal to 1 hour and less
                        than or equal to 48 hours.
                    minute_offset (int):
                        Optional. The number of minutes after the
                        hour (in UTC) to run the query. Must be greater
                        than or equal to 0 minutes and less than or
                        equal to 59 minutes.  If left unspecified, then
                        an arbitrary offset is used.

                        This field is a member of `oneof`_ ``_minute_offset``.
                """

                periodicity: int = proto.Field(
                    proto.INT32,
                    number=1,
                )
                minute_offset: int = proto.Field(
                    proto.INT32,
                    number=2,
                    optional=True,
                )

            class Daily(proto.Message):
                r"""Used to schedule the query to run every so many days.

                Attributes:
                    periodicity (int):
                        Required. The number of days between runs.
                        Must be greater than or equal to 1 day and less
                        than or equal to 31 days.
                    execution_time (google.type.timeofday_pb2.TimeOfDay):
                        Optional. The time of day (in UTC) at which
                        the query should run. If left unspecified, the
                        server picks an arbitrary time of day and runs
                        the query at the same time each day.
                """

                periodicity: int = proto.Field(
                    proto.INT32,
                    number=1,
                )
                execution_time: timeofday_pb2.TimeOfDay = proto.Field(
                    proto.MESSAGE,
                    number=2,
                    message=timeofday_pb2.TimeOfDay,
                )

            class RowCountTest(proto.Message):
                r"""A test that checks if the number of rows in the result set
                violates some threshold.

                Attributes:
                    comparison (google.cloud.monitoring_v3.types.ComparisonType):
                        Required. The comparison to apply between the
                        number of rows returned by the query and the
                        threshold.
                    threshold (int):
                        Required. The value against which to compare
                        the row count.
                """

                comparison: common.ComparisonType = proto.Field(
                    proto.ENUM,
                    number=1,
                    enum=common.ComparisonType,
                )
                threshold: int = proto.Field(
                    proto.INT64,
                    number=2,
                )

            class BooleanTest(proto.Message):
                r"""A test that uses an alerting result in a boolean column
                produced by the SQL query.

                Attributes:
                    column (str):
                        Required. The name of the column containing
                        the boolean value. If the value in a row is
                        NULL, that row is ignored.
                """

                column: str = proto.Field(
                    proto.STRING,
                    number=1,
                )

            query: str = proto.Field(
                proto.STRING,
                number=1,
            )
            minutes: "AlertPolicy.Condition.SqlCondition.Minutes" = proto.Field(
                proto.MESSAGE,
                number=2,
                oneof="schedule",
                message="AlertPolicy.Condition.SqlCondition.Minutes",
            )
            hourly: "AlertPolicy.Condition.SqlCondition.Hourly" = proto.Field(
                proto.MESSAGE,
                number=3,
                oneof="schedule",
                message="AlertPolicy.Condition.SqlCondition.Hourly",
            )
            daily: "AlertPolicy.Condition.SqlCondition.Daily" = proto.Field(
                proto.MESSAGE,
                number=4,
                oneof="schedule",
                message="AlertPolicy.Condition.SqlCondition.Daily",
            )
            row_count_test: "AlertPolicy.Condition.SqlCondition.RowCountTest" = (
                proto.Field(
                    proto.MESSAGE,
                    number=5,
                    oneof="evaluate",
                    message="AlertPolicy.Condition.SqlCondition.RowCountTest",
                )
            )
            boolean_test: "AlertPolicy.Condition.SqlCondition.BooleanTest" = (
                proto.Field(
                    proto.MESSAGE,
                    number=6,
                    oneof="evaluate",
                    message="AlertPolicy.Condition.SqlCondition.BooleanTest",
                )
            )

        name: str = proto.Field(
            proto.STRING,
            number=12,
        )
        display_name: str = proto.Field(
            proto.STRING,
            number=6,
        )
        condition_threshold: "AlertPolicy.Condition.MetricThreshold" = proto.Field(
            proto.MESSAGE,
            number=1,
            oneof="condition",
            message="AlertPolicy.Condition.MetricThreshold",
        )
        condition_absent: "AlertPolicy.Condition.MetricAbsence" = proto.Field(
            proto.MESSAGE,
            number=2,
            oneof="condition",
            message="AlertPolicy.Condition.MetricAbsence",
        )
        condition_matched_log: "AlertPolicy.Condition.LogMatch" = proto.Field(
            proto.MESSAGE,
            number=20,
            oneof="condition",
            message="AlertPolicy.Condition.LogMatch",
        )
        condition_monitoring_query_language: "AlertPolicy.Condition.MonitoringQueryLanguageCondition" = proto.Field(
            proto.MESSAGE,
            number=19,
            oneof="condition",
            message="AlertPolicy.Condition.MonitoringQueryLanguageCondition",
        )
        condition_prometheus_query_language: "AlertPolicy.Condition.PrometheusQueryLanguageCondition" = proto.Field(
            proto.MESSAGE,
            number=21,
            oneof="condition",
            message="AlertPolicy.Condition.PrometheusQueryLanguageCondition",
        )
        condition_sql: "AlertPolicy.Condition.SqlCondition" = proto.Field(
            proto.MESSAGE,
            number=22,
            oneof="condition",
            message="AlertPolicy.Condition.SqlCondition",
        )

    class AlertStrategy(proto.Message):
        r"""Control over how the notification channels in
        ``notification_channels`` are notified when this alert fires.

        Attributes:
            notification_rate_limit (google.cloud.monitoring_v3.types.AlertPolicy.AlertStrategy.NotificationRateLimit):
                Required for log-based alerting policies, i.e. policies with
                a ``LogMatch`` condition.

                This limit is not implemented for alerting policies that do
                not have a LogMatch condition.
            notification_prompts (MutableSequence[google.cloud.monitoring_v3.types.AlertPolicy.AlertStrategy.NotificationPrompt]):
                For log-based alert policies, the notification prompts is
                always [OPENED]. For non log-based alert policies, the
                notification prompts can be [OPENED] or [OPENED, CLOSED].
            auto_close (google.protobuf.duration_pb2.Duration):
                If an alerting policy that was active has no
                data for this long, any open incidents will
                close
            notification_channel_strategy (MutableSequence[google.cloud.monitoring_v3.types.AlertPolicy.AlertStrategy.NotificationChannelStrategy]):
                Control how notifications will be sent out,
                on a per-channel basis.
        """

        class NotificationPrompt(proto.Enum):
            r"""Control when notifications will be sent out.

            Values:
                NOTIFICATION_PROMPT_UNSPECIFIED (0):
                    No strategy specified. Treated as error.
                OPENED (1):
                    Notify when an incident is opened.
                CLOSED (3):
                    Notify when an incident is closed.
            """
            NOTIFICATION_PROMPT_UNSPECIFIED = 0
            OPENED = 1
            CLOSED = 3

        class NotificationRateLimit(proto.Message):
            r"""Control over the rate of notifications sent to this alerting
            policy's notification channels.

            Attributes:
                period (google.protobuf.duration_pb2.Duration):
                    Not more than one notification per ``period``.
            """

            period: duration_pb2.Duration = proto.Field(
                proto.MESSAGE,
                number=1,
                message=duration_pb2.Duration,
            )

        class NotificationChannelStrategy(proto.Message):
            r"""Control over how the notification channels in
            ``notification_channels`` are notified when this alert fires, on a
            per-channel basis.

            Attributes:
                notification_channel_names (MutableSequence[str]):
                    The full REST resource name for the notification channels
                    that these settings apply to. Each of these correspond to
                    the name field in one of the NotificationChannel objects
                    referenced in the notification_channels field of this
                    AlertPolicy. The format is:

                    ::

                        projects/[PROJECT_ID_OR_NUMBER]/notificationChannels/[CHANNEL_ID]
                renotify_interval (google.protobuf.duration_pb2.Duration):
                    The frequency at which to send reminder
                    notifications for open incidents.
            """

            notification_channel_names: MutableSequence[str] = proto.RepeatedField(
                proto.STRING,
                number=1,
            )
            renotify_interval: duration_pb2.Duration = proto.Field(
                proto.MESSAGE,
                number=2,
                message=duration_pb2.Duration,
            )

        notification_rate_limit: "AlertPolicy.AlertStrategy.NotificationRateLimit" = (
            proto.Field(
                proto.MESSAGE,
                number=1,
                message="AlertPolicy.AlertStrategy.NotificationRateLimit",
            )
        )
        notification_prompts: MutableSequence[
            "AlertPolicy.AlertStrategy.NotificationPrompt"
        ] = proto.RepeatedField(
            proto.ENUM,
            number=2,
            enum="AlertPolicy.AlertStrategy.NotificationPrompt",
        )
        auto_close: duration_pb2.Duration = proto.Field(
            proto.MESSAGE,
            number=3,
            message=duration_pb2.Duration,
        )
        notification_channel_strategy: MutableSequence[
            "AlertPolicy.AlertStrategy.NotificationChannelStrategy"
        ] = proto.RepeatedField(
            proto.MESSAGE,
            number=4,
            message="AlertPolicy.AlertStrategy.NotificationChannelStrategy",
        )

    name: str = proto.Field(
        proto.STRING,
        number=1,
    )
    display_name: str = proto.Field(
        proto.STRING,
        number=2,
    )
    documentation: Documentation = proto.Field(
        proto.MESSAGE,
        number=13,
        message=Documentation,
    )
    user_labels: MutableMapping[str, str] = proto.MapField(
        proto.STRING,
        proto.STRING,
        number=16,
    )
    conditions: MutableSequence[Condition] = proto.RepeatedField(
        proto.MESSAGE,
        number=12,
        message=Condition,
    )
    combiner: ConditionCombinerType = proto.Field(
        proto.ENUM,
        number=6,
        enum=ConditionCombinerType,
    )
    enabled: wrappers_pb2.BoolValue = proto.Field(
        proto.MESSAGE,
        number=17,
        message=wrappers_pb2.BoolValue,
    )
    validity: status_pb2.Status = proto.Field(
        proto.MESSAGE,
        number=18,
        message=status_pb2.Status,
    )
    notification_channels: MutableSequence[str] = proto.RepeatedField(
        proto.STRING,
        number=14,
    )
    creation_record: gm_mutation_record.MutationRecord = proto.Field(
        proto.MESSAGE,
        number=10,
        message=gm_mutation_record.MutationRecord,
    )
    mutation_record: gm_mutation_record.MutationRecord = proto.Field(
        proto.MESSAGE,
        number=11,
        message=gm_mutation_record.MutationRecord,
    )
    alert_strategy: AlertStrategy = proto.Field(
        proto.MESSAGE,
        number=21,
        message=AlertStrategy,
    )
    severity: Severity = proto.Field(
        proto.ENUM,
        number=22,
        enum=Severity,
    )


__all__ = tuple(sorted(__protobuf__.manifest))



# -*- coding: utf-8 -*-
# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import os

# try/except added for compatibility with python < 3.8
try:
    from unittest import mock
    from unittest.mock import AsyncMock  # pragma: NO COVER
except ImportError:  # pragma: NO COVER
    import mock

import json
import math

from google.api_core import api_core_version
import grpc
from grpc.experimental import aio
from proto.marshal.rules import wrappers
from proto.marshal.rules.dates import DurationRule, TimestampRule
import pytest

try:
    from google.auth.aio import credentials as ga_credentials_async

    HAS_GOOGLE_AUTH_AIO = True
except ImportError:  # pragma: NO COVER
    HAS_GOOGLE_AUTH_AIO = False

from google.api_core import gapic_v1, grpc_helpers, grpc_helpers_async, path_template
from google.api_core import client_options
from google.api_core import exceptions as core_exceptions
from google.api_core import retry as retries
import google.auth
from google.auth import credentials as ga_credentials
from google.auth.exceptions import MutualTLSChannelError
from google.longrunning import operations_pb2  # type: ignore
from google.oauth2 import service_account

from google.cloud.monitoring_v3.services.query_service import (
    QueryServiceAsyncClient,
    QueryServiceClient,
    pagers,
    transports,
)
from google.cloud.monitoring_v3.types import metric, metric_service

CRED_INFO_JSON = {
    "credential_source": "/path/to/file",
    "credential_type": "service account credentials",
    "principal": "service-account@example.com",
}
CRED_INFO_STRING = json.dumps(CRED_INFO_JSON)


async def mock_async_gen(data, chunk_size=1):
    for i in range(0, len(data)):  # pragma: NO COVER
        chunk = data[i : i + chunk_size]
        yield chunk.encode("utf-8")


def client_cert_source_callback():
    return b"cert bytes", b"key bytes"


# TODO: use async auth anon credentials by default once the minimum version of google-auth is upgraded.
# See related issue: https://github.com/googleapis/gapic-generator-python/issues/2107.
def async_anonymous_credentials():
    if HAS_GOOGLE_AUTH_AIO:
        return ga_credentials_async.AnonymousCredentials()
    return ga_credentials.AnonymousCredentials()


# If default endpoint is localhost, then default mtls endpoint will be the same.
# This method modifies the default endpoint so the client can produce a different
# mtls endpoint for endpoint testing purposes.
def modify_default_endpoint(client):
    return (
        "foo.googleapis.com"
        if ("localhost" in client.DEFAULT_ENDPOINT)
        else client.DEFAULT_ENDPOINT
    )


# If default endpoint template is localhost, then default mtls endpoint will be the same.
# This method modifies the default endpoint template so the client can produce a different
# mtls endpoint for endpoint testing purposes.
def modify_default_endpoint_template(client):
    return (
        "test.{UNIVERSE_DOMAIN}"
        if ("localhost" in client._DEFAULT_ENDPOINT_TEMPLATE)
        else client._DEFAULT_ENDPOINT_TEMPLATE
    )


def test__get_default_mtls_endpoint():
    api_endpoint = "example.googleapis.com"
    api_mtls_endpoint = "example.mtls.googleapis.com"
    sandbox_endpoint = "example.sandbox.googleapis.com"
    sandbox_mtls_endpoint = "example.mtls.sandbox.googleapis.com"
    non_googleapi = "api.example.com"

    assert QueryServiceClient._get_default_mtls_endpoint(None) is None
    assert (
        QueryServiceClient._get_default_mtls_endpoint(api_endpoint) == api_mtls_endpoint
    )
    assert (
        QueryServiceClient._get_default_mtls_endpoint(api_mtls_endpoint)
        == api_mtls_endpoint
    )
    assert (
        QueryServiceClient._get_default_mtls_endpoint(sandbox_endpoint)
        == sandbox_mtls_endpoint
    )
    assert (
        QueryServiceClient._get_default_mtls_endpoint(sandbox_mtls_endpoint)
        == sandbox_mtls_endpoint
    )
    assert QueryServiceClient._get_default_mtls_endpoint(non_googleapi) == non_googleapi


def test__read_environment_variables():
    assert QueryServiceClient._read_environment_variables() == (False, "auto", None)

    with mock.patch.dict(os.environ, {"GOOGLE_API_USE_CLIENT_CERTIFICATE": "true"}):
        assert QueryServiceClient._read_environment_variables() == (True, "auto", None)

    with mock.patch.dict(os.environ, {"GOOGLE_API_USE_CLIENT_CERTIFICATE": "false"}):
        assert QueryServiceClient._read_environment_variables() == (False, "auto", None)

    with mock.patch.dict(
        os.environ, {"GOOGLE_API_USE_CLIENT_CERTIFICATE": "Unsupported"}
    ):
        if not hasattr(google.auth.transport.mtls, "should_use_client_cert"):
            with pytest.raises(ValueError) as excinfo:
                QueryServiceClient._read_environment_variables()
            assert (
                str(excinfo.value)
                == "Environment variable `GOOGLE_API_USE_CLIENT_CERTIFICATE` must be either `true` or `false`"
            )
        else:
            assert QueryServiceClient._read_environment_variables() == (
                False,
                "auto",
                None,
            )

    with mock.patch.dict(os.environ, {"GOOGLE_API_USE_MTLS_ENDPOINT": "never"}):
        assert QueryServiceClient._read_environment_variables() == (
            False,
            "never",
            None,
        )

    with mock.patch.dict(os.environ, {"GOOGLE_API_USE_MTLS_ENDPOINT": "always"}):
        assert QueryServiceClient._read_environment_variables() == (
            False,
            "always",
            None,
        )

    with mock.patch.dict(os.environ, {"GOOGLE_API_USE_MTLS_ENDPOINT": "auto"}):
        assert QueryServiceClient._read_environment_variables() == (False, "auto", None)

    with mock.patch.dict(os.environ, {"GOOGLE_API_USE_MTLS_ENDPOINT": "Unsupported"}):
        with pytest.raises(MutualTLSChannelError) as excinfo:
            QueryServiceClient._read_environment_variables()
    assert (
        str(excinfo.value)
        == "Environment variable `GOOGLE_API_USE_MTLS_ENDPOINT` must be `never`, `auto` or `always`"
    )

    with mock.patch.dict(os.environ, {"GOOGLE_CLOUD_UNIVERSE_DOMAIN": "foo.com"}):
        assert QueryServiceClient._read_environment_variables() == (
            False,
            "auto",
            "foo.com",
        )


def test_use_client_cert_effective():
    # Test case 1: Test when `should_use_client_cert` returns True.
    # We mock the `should_use_client_cert` function to simulate a scenario where
    # the google-auth library supports automatic mTLS and determines that a
    # client certificate should be used.
    if hasattr(google.auth.transport.mtls, "should_use_client_cert"):
        with mock.patch(
            "google.auth.transport.mtls.should_use_client_cert", return_value=True
        ):
            assert QueryServiceClient._use_client_cert_effective() is True

    # Test case 2: Test when `should_use_client_cert` returns False.
    # We mock the `should_use_client_cert` function to simulate a scenario where
    # the google-auth library supports automatic mTLS and determines that a
    # client certificate should NOT be used.
    if hasattr(google.auth.transport.mtls, "should_use_client_cert"):
        with mock.patch(
            "google.auth.transport.mtls.should_use_client_cert", return_value=False
        ):
            assert QueryServiceClient._use_client_cert_effective() is False

    # Test case 3: Test when `should_use_client_cert` is unavailable and the
    # `GOOGLE_API_USE_CLIENT_CERTIFICATE` environment variable is set to "true".
    if not hasattr(google.auth.transport.mtls, "should_use_client_cert"):
        with mock.patch.dict(os.environ, {"GOOGLE_API_USE_CLIENT_CERTIFICATE": "true"}):
            assert QueryServiceClient._use_client_cert_effective() is True

    # Test case 4: Test when `should_use_client_cert` is unavailable and the
    # `GOOGLE_API_USE_CLIENT_CERTIFICATE` environment variable is set to "false".
    if not hasattr(google.auth.transport.mtls, "should_use_client_cert"):
        with mock.patch.dict(
            os.environ, {"GOOGLE_API_USE_CLIENT_CERTIFICATE": "false"}
        ):
            assert QueryServiceClient._use_client_cert_effective() is False

    # Test case 5: Test when `should_use_client_cert` is unavailable and the
    # `GOOGLE_API_USE_CLIENT_CERTIFICATE` environment variable is set to "True".
    if not hasattr(google.auth.transport.mtls, "should_use_client_cert"):
        with mock.patch.dict(os.environ, {"GOOGLE_API_USE_CLIENT_CERTIFICATE": "True"}):
            assert QueryServiceClient._use_client_cert_effective() is True

    # Test case 6: Test when `should_use_client_cert` is unavailable and the
    # `GOOGLE_API_USE_CLIENT_CERTIFICATE` environment variable is set to "False".
    if not hasattr(google.auth.transport.mtls, "should_use_client_cert"):
        with mock.patch.dict(
            os.environ, {"GOOGLE_API_USE_CLIENT_CERTIFICATE": "False"}
        ):
            assert QueryServiceClient._use_client_cert_effective() is False

    # Test case 7: Test when `should_use_client_cert` is unavailable and the
    # `GOOGLE_API_USE_CLIENT_CERTIFICATE` environment variable is set to "TRUE".
    if not hasattr(google.auth.transport.mtls, "should_use_client_cert"):
        with mock.patch.dict(os.environ, {"GOOGLE_API_USE_CLIENT_CERTIFICATE": "TRUE"}):
            assert QueryServiceClient._use_client_cert_effective() is True

    # Test case 8: Test when `should_use_client_cert` is unavailable and the
    # `GOOGLE_API_USE_CLIENT_CERTIFICATE` environment variable is set to "FALSE".
    if not hasattr(google.auth.transport.mtls, "should_use_client_cert"):
        with mock.patch.dict(
            os.environ, {"GOOGLE_API_USE_CLIENT_CERTIFICATE": "FALSE"}
        ):
            assert QueryServiceClient._use_client_cert_effective() is False

    # Test case 9: Test when `should_use_client_cert` is unavailable and the
    # `GOOGLE_API_USE_CLIENT_CERTIFICATE` environment variable is not set.
    # In this case, the method should return False, which is the default value.
    if not hasattr(google.auth.transport.mtls, "should_use_client_cert"):
        with mock.patch.dict(os.environ, clear=True):
            assert QueryServiceClient._use_client_cert_effective() is False

    # Test case 10: Test when `should_use_client_cert` is unavailable and the
    # `GOOGLE_API_USE_CLIENT_CERTIFICATE` environment variable is set to an invalid value.
    # The method should raise a ValueError as the environment variable must be either
    # "true" or "false".
    if not hasattr(google.auth.transport.mtls, "should_use_client_cert"):
        with mock.patch.dict(
            os.environ, {"GOOGLE_API_USE_CLIENT_CERTIFICATE": "unsupported"}
        ):
            with pytest.raises(ValueError):
                QueryServiceClient._use_client_cert_effective()

    # Test case 11: Test when `should_use_client_cert` is available and the
    # `GOOGLE_API_USE_CLIENT_CERTIFICATE` environment variable is set to an invalid value.
    # The method should return False as the environment variable is set to an invalid value.
    if hasattr(google.auth.transport.mtls, "should_use_client_cert"):
        with mock.patch.dict(
            os.environ, {"GOOGLE_API_USE_CLIENT_CERTIFICATE": "unsupported"}
        ):
            assert QueryServiceClient._use_client_cert_effective() is False

    # Test case 12: Test when `should_use_client_cert` is available and the
    # `GOOGLE_API_USE_CLIENT_CERTIFICATE` environment variable is unset. Also,
    # the GOOGLE_API_CONFIG environment variable is unset.
    if hasattr(google.auth.transport.mtls, "should_use_client_cert"):
        with mock.patch.dict(os.environ, {"GOOGLE_API_USE_CLIENT_CERTIFICATE": ""}):
            with mock.patch.dict(os.environ, {"GOOGLE_API_CERTIFICATE_CONFIG": ""}):
                assert QueryServiceClient._use_client_cert_effective() is False


def test__get_client_cert_source():
    mock_provided_cert_source = mock.Mock()
    mock_default_cert_source = mock.Mock()

    assert QueryServiceClient._get_client_cert_source(None, False) is None
    assert (
        QueryServiceClient._get_client_cert_source(mock_provided_cert_source, False)
        is None
    )
    assert (
        QueryServiceClient._get_client_cert_source(mock_provided_cert_source, True)
        == mock_provided_cert_source
    )

    with mock.patch(
        "google.auth.transport.mtls.has_default_client_cert_source", return_value=True
    ):
        with mock.patch(
            "google.auth.transport.mtls.default_client_cert_source",
            return_value=mock_default_cert_source,
        ):
            assert (
                QueryServiceClient._get_client_cert_source(None, True)
                is mock_default_cert_source
            )
            assert (
                QueryServiceClient._get_client_cert_source(
                    mock_provided_cert_source, "true"
                )
                is mock_provided_cert_source
            )


@mock.patch.object(
    QueryServiceClient,
    "_DEFAULT_ENDPOINT_TEMPLATE",
    modify_default_endpoint_template(QueryServiceClient),
)
@mock.patch.object(
    QueryServiceAsyncClient,
    "_DEFAULT_ENDPOINT_TEMPLATE",
    modify_default_endpoint_template(QueryServiceAsyncClient),
)
def test__get_api_endpoint():
    api_override = "foo.com"
    mock_client_cert_source = mock.Mock()
    default_universe = QueryServiceClient._DEFAULT_UNIVERSE
    default_endpoint = QueryServiceClient._DEFAULT_ENDPOINT_TEMPLATE.format(
        UNIVERSE_DOMAIN=default_universe
    )
    mock_universe = "bar.com"
    mock_endpoint = QueryServiceClient._DEFAULT_ENDPOINT_TEMPLATE.format(
        UNIVERSE_DOMAIN=mock_universe
    )

    assert (
        QueryServiceClient._get_api_endpoint(
            api_override, mock_client_cert_source, default_universe, "always"
        )
        == api_override
    )
    assert (
        QueryServiceClient._get_api_endpoint(
            None, mock_client_cert_source, default_universe, "auto"
        )
        == QueryServiceClient.DEFAULT_MTLS_ENDPOINT
    )
    assert (
        QueryServiceClient._get_api_endpoint(None, None, default_universe, "auto")
        == default_endpoint
    )
    assert (
        QueryServiceClient._get_api_endpoint(None, None, default_universe, "always")
        == QueryServiceClient.DEFAULT_MTLS_ENDPOINT
    )
    assert (
        QueryServiceClient._get_api_endpoint(
            None, mock_client_cert_source, default_universe, "always"
        )
        == QueryServiceClient.DEFAULT_MTLS_ENDPOINT
    )
    assert (
        QueryServiceClient._get_api_endpoint(None, None, mock_universe, "never")
        == mock_endpoint
    )
    assert (
        QueryServiceClient._get_api_endpoint(None, None, default_universe, "never")
        == default_endpoint
    )

    with pytest.raises(MutualTLSChannelError) as excinfo:
        QueryServiceClient._get_api_endpoint(
            None, mock_client_cert_source, mock_universe, "auto"
        )
    assert (
        str(excinfo.value)
        == "mTLS is not supported in any universe other than googleapis.com."
    )


def test__get_universe_domain():
    client_universe_domain = "foo.com"
    universe_domain_env = "bar.com"

    assert (
        QueryServiceClient._get_universe_domain(
            client_universe_domain, universe_domain_env
        )
        == client_universe_domain
    )
    assert (
        QueryServiceClient._get_universe_domain(None, universe_domain_env)
        == universe_domain_env
    )
    assert (
        QueryServiceClient._get_universe_domain(None, None)
        == QueryServiceClient._DEFAULT_UNIVERSE
    )

    with pytest.raises(ValueError) as excinfo:
        QueryServiceClient._get_universe_domain("", None)
    assert str(excinfo.value) == "Universe Domain cannot be an empty string."


@pytest.mark.parametrize(
    "error_code,cred_info_json,show_cred_info",
    [
        (401, CRED_INFO_JSON, True),
        (403, CRED_INFO_JSON, True),
        (404, CRED_INFO_JSON, True),
        (500, CRED_INFO_JSON, False),
        (401, None, False),
        (403, None, False),
        (404, None, False),
        (500, None, False),
    ],
)
def test__add_cred_info_for_auth_errors(error_code, cred_info_json, show_cred_info):
    cred = mock.Mock(["get_cred_info"])
    cred.get_cred_info = mock.Mock(return_value=cred_info_json)
    client = QueryServiceClient(credentials=cred)
    client._transport._credentials = cred

    error = core_exceptions.GoogleAPICallError("message", details=["foo"])
    error.code = error_code

    client._add_cred_info_for_auth_errors(error)
    if show_cred_info:
        assert error.details == ["foo", CRED_INFO_STRING]
    else:
        assert error.details == ["foo"]


@pytest.mark.parametrize("error_code", [401, 403, 404, 500])
def test__add_cred_info_for_auth_errors_no_get_cred_info(error_code):
    cred = mock.Mock([])
    assert not hasattr(cred, "get_cred_info")
    client = QueryServiceClient(credentials=cred)
    client._transport._credentials = cred

    error = core_exceptions.GoogleAPICallError("message", details=[])
    error.code = error_code

    client._add_cred_info_for_auth_errors(error)
    assert error.details == []


@pytest.mark.parametrize(
    "client_class,transport_name",
    [
        (QueryServiceClient, "grpc"),
        (QueryServiceAsyncClient, "grpc_asyncio"),
    ],
)
def test_query_service_client_from_service_account_info(client_class, transport_name):
    creds = ga_credentials.AnonymousCredentials()
    with mock.patch.object(
        service_account.Credentials, "from_service_account_info"
    ) as factory:
        factory.return_value = creds
        info = {"valid": True}
        client = client_class.from_service_account_info(info, transport=transport_name)
        assert client.transport._credentials == creds
        assert isinstance(client, client_class)

        assert client.transport._host == ("monitoring.googleapis.com:443")


@pytest.mark.parametrize(
    "transport_class,transport_name",
    [
        (transports.QueryServiceGrpcTransport, "grpc"),
        (transports.QueryServiceGrpcAsyncIOTransport, "grpc_asyncio"),
    ],
)
def test_query_service_client_service_account_always_use_jwt(
    transport_class, transport_name
):
    with mock.patch.object(
        service_account.Credentials, "with_always_use_jwt_access", create=True
    ) as use_jwt:
        creds = service_account.Credentials(None, None, None)
        transport = transport_class(credentials=creds, always_use_jwt_access=True)
        use_jwt.assert_called_once_with(True)

    with mock.patch.object(
        service_account.Credentials, "with_always_use_jwt_access", create=True
    ) as use_jwt:
        creds = service_account.Credentials(None, None, None)
        transport = transport_class(credentials=creds, always_use_jwt_access=False)
        use_jwt.assert_not_called()


@pytest.mark.parametrize(
    "client_class,transport_name",
    [
        (QueryServiceClient, "grpc"),
        (QueryServiceAsyncClient, "grpc_asyncio"),
    ],
)
def test_query_service_client_from_service_account_file(client_class, transport_name):
    creds = ga_credentials.AnonymousCredentials()
    with mock.patch.object(
        service_account.Credentials, "from_service_account_file"
    ) as factory:
        factory.return_value = creds
        client = client_class.from_service_account_file(
            "dummy/file/path.json", transport=transport_name
        )
        assert client.transport._credentials == creds
        assert isinstance(client, client_class)

        client = client_class.from_service_account_json(
            "dummy/file/path.json", transport=transport_name
        )
        assert client.transport._credentials == creds
        assert isinstance(client, client_class)

        assert client.transport._host == ("monitoring.googleapis.com:443")


def test_query_service_client_get_transport_class():
    transport = QueryServiceClient.get_transport_class()
    available_transports = [
        transports.QueryServiceGrpcTransport,
    ]
    assert transport in available_transports

    transport = QueryServiceClient.get_transport_class("grpc")
    assert transport == transports.QueryServiceGrpcTransport


@pytest.mark.parametrize(
    "client_class,transport_class,transport_name",
    [
        (QueryServiceClient, transports.QueryServiceGrpcTransport, "grpc"),
        (
            QueryServiceAsyncClient,
            transports.QueryServiceGrpcAsyncIOTransport,
            "grpc_asyncio",
        ),
    ],
)
@mock.patch.object(
    QueryServiceClient,
    "_DEFAULT_ENDPOINT_TEMPLATE",
    modify_default_endpoint_template(QueryServiceClient),
)
@mock.patch.object(
    QueryServiceAsyncClient,
    "_DEFAULT_ENDPOINT_TEMPLATE",
    modify_default_endpoint_template(QueryServiceAsyncClient),
)
def test_query_service_client_client_options(
    client_class, transport_class, transport_name
):
    # Check that if channel is provided we won't create a new one.
    with mock.patch.object(QueryServiceClient, "get_transport_class") as gtc:
        transport = transport_class(credentials=ga_credentials.AnonymousCredentials())
        client = client_class(transport=transport)
        gtc.assert_not_called()

    # Check that if channel is provided via str we will create a new one.
    with mock.patch.object(QueryServiceClient, "get_transport_class") as gtc:
        client = client_class(transport=transport_name)
        gtc.assert_called()

    # Check the case api_endpoint is provided.
    options = client_options.ClientOptions(api_endpoint="squid.clam.whelk")
    with mock.patch.object(transport_class, "__init__") as patched:
        patched.return_value = None
        client = client_class(transport=transport_name, client_options=options)
        patched.assert_called_once_with(
            credentials=None,
            credentials_file=None,
            host="squid.clam.whelk",
            scopes=None,
            client_cert_source_for_mtls=None,
            quota_project_id=None,
            client_info=transports.base.DEFAULT_CLIENT_INFO,
            always_use_jwt_access=True,
            api_audience=None,
        )

    # Check the case api_endpoint is not provided and GOOGLE_API_USE_MTLS_ENDPOINT is
    # "never".
    with mock.patch.dict(os.environ, {"GOOGLE_API_USE_MTLS_ENDPOINT": "never"}):
        with mock.patch.object(transport_class, "__init__") as patched:
            patched.return_value = None
            client = client_class(transport=transport_name)
            patched.assert_called_once_with(
                credentials=None,
                credentials_file=None,
                host=client._DEFAULT_ENDPOINT_TEMPLATE.format(
                    UNIVERSE_DOMAIN=client._DEFAULT_UNIVERSE
                ),
                scopes=None,
                client_cert_source_for_mtls=None,
                quota_project_id=None,
                client_info=transports.base.DEFAULT_CLIENT_INFO,
                always_use_jwt_access=True,
                api_audience=None,
            )

    # Check the case api_endpoint is not provided and GOOGLE_API_USE_MTLS_ENDPOINT is
    # "always".
    with mock.patch.dict(os.environ, {"GOOGLE_API_USE_MTLS_ENDPOINT": "always"}):
        with mock.patch.object(transport_class, "__init__") as patched:
            patched.return_value = None
            client = client_class(transport=transport_name)
            patched.assert_called_once_with(
                credentials=None,
                credentials_file=None,
                host=client.DEFAULT_MTLS_ENDPOINT,
                scopes=None,
                client_cert_source_for_mtls=None,
                quota_project_id=None,
                client_info=transports.base.DEFAULT_CLIENT_INFO,
                always_use_jwt_access=True,
                api_audience=None,
            )

    # Check the case api_endpoint is not provided and GOOGLE_API_USE_MTLS_ENDPOINT has
    # unsupported value.
    with mock.patch.dict(os.environ, {"GOOGLE_API_USE_MTLS_ENDPOINT": "Unsupported"}):
        with pytest.raises(MutualTLSChannelError) as excinfo:
            client = client_class(transport=transport_name)
    assert (
        str(excinfo.value)
        == "Environment variable `GOOGLE_API_USE_MTLS_ENDPOINT` must be `never`, `auto` or `always`"
    )

    # Check the case quota_project_id is provided
    options = client_options.ClientOptions(quota_project_id="octopus")
    with mock.patch.object(transport_class, "__init__") as patched:
        patched.return_value = None
        client = client_class(client_options=options, transport=transport_name)
        patched.assert_called_once_with(
            credentials=None,
            credentials_file=None,
            host=client._DEFAULT_ENDPOINT_TEMPLATE.format(
                UNIVERSE_DOMAIN=client._DEFAULT_UNIVERSE
            ),
            scopes=None,
            client_cert_source_for_mtls=None,
            quota_project_id="octopus",
            client_info=transports.base.DEFAULT_CLIENT_INFO,
            always_use_jwt_access=True,
            api_audience=None,
        )
    # Check the case api_endpoint is provided
    options = client_options.ClientOptions(
        api_audience="https://language.googleapis.com"
    )
    with mock.patch.object(transport_class, "__init__") as patched:
        patched.return_value = None
        client = client_class(client_options=options, transport=transport_name)
        patched.assert_called_once_with(
            credentials=None,
            credentials_file=None,
            host=client._DEFAULT_ENDPOINT_TEMPLATE.format(
                UNIVERSE_DOMAIN=client._DEFAULT_UNIVERSE
            ),
            scopes=None,
            client_cert_source_for_mtls=None,
            quota_project_id=None,
            client_info=transports.base.DEFAULT_CLIENT_INFO,
            always_use_jwt_access=True,
            api_audience="https://language.googleapis.com",
        )


@pytest.mark.parametrize(
    "client_class,transport_class,transport_name,use_client_cert_env",
    [
        (QueryServiceClient, transports.QueryServiceGrpcTransport, "grpc", "true"),
        (
            QueryServiceAsyncClient,
            transports.QueryServiceGrpcAsyncIOTransport,
            "grpc_asyncio",
            "true",
        ),
        (QueryServiceClient, transports.QueryServiceGrpcTransport, "grpc", "false"),
        (
            QueryServiceAsyncClient,
            transports.QueryServiceGrpcAsyncIOTransport,
            "grpc_asyncio",
            "false",
        ),
    ],
)
@mock.patch.object(
    QueryServiceClient,
    "_DEFAULT_ENDPOINT_TEMPLATE",
    modify_default_endpoint_template(QueryServiceClient),
)
@mock.patch.object(
    QueryServiceAsyncClient,
    "_DEFAULT_ENDPOINT_TEMPLATE",
    modify_default_endpoint_template(QueryServiceAsyncClient),
)
@mock.patch.dict(os.environ, {"GOOGLE_API_USE_MTLS_ENDPOINT": "auto"})
def test_query_service_client_mtls_env_auto(
    client_class, transport_class, transport_name, use_client_cert_env
):
    # This tests the endpoint autoswitch behavior. Endpoint is autoswitched to the default
    # mtls endpoint, if GOOGLE_API_USE_CLIENT_CERTIFICATE is "true" and client cert exists.

    # Check the case client_cert_source is provided. Whether client cert is used depends on
    # GOOGLE_API_USE_CLIENT_CERTIFICATE value.
    with mock.patch.dict(
        os.environ, {"GOOGLE_API_USE_CLIENT_CERTIFICATE": use_client_cert_env}
    ):
        options = client_options.ClientOptions(
            client_cert_source=client_cert_source_callback
        )
        with mock.patch.object(transport_class, "__init__") as patched:
            patched.return_value = None
            client = client_class(client_options=options, transport=transport_name)

            if use_client_cert_env == "false":
                expected_client_cert_source = None
                expected_host = client._DEFAULT_ENDPOINT_TEMPLATE.format(
                    UNIVERSE_DOMAIN=client._DEFAULT_UNIVERSE
                )
            else:
                expected_client_cert_source = client_cert_source_callback
                expected_host = client.DEFAULT_MTLS_ENDPOINT

            patched.assert_called_once_with(
                credentials=None,
                credentials_file=None,
                host=expected_host,
                scopes=None,
                client_cert_source_for_mtls=expected_client_cert_source,
                quota_project_id=None,
                client_info=transports.base.DEFAULT_CLIENT_INFO,
                always_use_jwt_access=True,
                api_audience=None,
            )

    # Check the case ADC client cert is provided. Whether client cert is used depends on
    # GOOGLE_API_USE_CLIENT_CERTIFICATE value.
    with mock.patch.dict(
        os.environ, {"GOOGLE_API_USE_CLIENT_CERTIFICATE": use_client_cert_env}
    ):
        with mock.patch.object(transport_class, "__init__") as patched:
            with mock.patch(
                "google.auth.transport.mtls.has_default_client_cert_source",
                return_value=True,
            ):
                with mock.patch(
                    "google.auth.transport.mtls.default_client_cert_source",
                    return_value=client_cert_source_callback,
                ):
                    if use_client_cert_env == "false":
                        expected_host = client._DEFAULT_ENDPOINT_TEMPLATE.format(
                            UNIVERSE_DOMAIN=client._DEFAULT_UNIVERSE
                        )
                        expected_client_cert_source = None
                    else:
                        expected_host = client.DEFAULT_MTLS_ENDPOINT
                        expected_client_cert_source = client_cert_source_callback

                    patched.return_value = None
                    client = client_class(transport=transport_name)
                    patched.assert_called_once_with(
                        credentials=None,
                        credentials_file=None,
                        host=expected_host,
                        scopes=None,
                        client_cert_source_for_mtls=expected_client_cert_source,
                        quota_project_id=None,
                        client_info=transports.base.DEFAULT_CLIENT_INFO,
                        always_use_jwt_access=True,
                        api_audience=None,
                    )

    # Check the case client_cert_source and ADC client cert are not provided.
    with mock.patch.dict(
        os.environ, {"GOOGLE_API_USE_CLIENT_CERTIFICATE": use_client_cert_env}
    ):
        with mock.patch.object(transport_class, "__init__") as patched:
            with mock.patch(
                "google.auth.transport.mtls.has_default_client_cert_source",
                return_value=False,
            ):
                patched.return_value = None
                client = client_class(transport=transport_name)
                patched.assert_called_once_with(
                    credentials=None,
                    credentials_file=None,
                    host=client._DEFAULT_ENDPOINT_TEMPLATE.format(
                        UNIVERSE_DOMAIN=client._DEFAULT_UNIVERSE
                    ),
                    scopes=None,
                    client_cert_source_for_mtls=None,
                    quota_project_id=None,
                    client_info=transports.base.DEFAULT_CLIENT_INFO,
                    always_use_jwt_access=True,
                    api_audience=None,
                )


@pytest.mark.parametrize("client_class", [QueryServiceClient, QueryServiceAsyncClient])
@mock.patch.object(
    QueryServiceClient, "DEFAULT_ENDPOINT", modify_default_endpoint(QueryServiceClient)
)
@mock.patch.object(
    QueryServiceAsyncClient,
    "DEFAULT_ENDPOINT",
    modify_default_endpoint(QueryServiceAsyncClient),
)
def test_query_service_client_get_mtls_endpoint_and_cert_source(client_class):
    mock_client_cert_source = mock.Mock()

    # Test the case GOOGLE_API_USE_CLIENT_CERTIFICATE is "true".
    with mock.patch.dict(os.environ, {"GOOGLE_API_USE_CLIENT_CERTIFICATE": "true"}):
        mock_api_endpoint = "foo"
        options = client_options.ClientOptions(
            client_cert_source=mock_client_cert_source, api_endpoint=mock_api_endpoint
        )
        api_endpoint, cert_source = client_class.get_mtls_endpoint_and_cert_source(
            options
        )
        assert api_endpoint == mock_api_endpoint
        assert cert_source == mock_client_cert_source

    # Test the case GOOGLE_API_USE_CLIENT_CERTIFICATE is "false".
    with mock.patch.dict(os.environ, {"GOOGLE_API_USE_CLIENT_CERTIFICATE": "false"}):
        mock_client_cert_source = mock.Mock()
        mock_api_endpoint = "foo"
        options = client_options.ClientOptions(
            client_cert_source=mock_client_cert_source, api_endpoint=mock_api_endpoint
        )
        api_endpoint, cert_source = client_class.get_mtls_endpoint_and_cert_source(
            options
        )
        assert api_endpoint == mock_api_endpoint
        assert cert_source is None

    # Test the case GOOGLE_API_USE_CLIENT_CERTIFICATE is "Unsupported".
    with mock.patch.dict(
        os.environ, {"GOOGLE_API_USE_CLIENT_CERTIFICATE": "Unsupported"}
    ):
        if hasattr(google.auth.transport.mtls, "should_use_client_cert"):
            mock_client_cert_source = mock.Mock()
            mock_api_endpoint = "foo"
            options = client_options.ClientOptions(
                client_cert_source=mock_client_cert_source,
                api_endpoint=mock_api_endpoint,
            )
            api_endpoint, cert_source = client_class.get_mtls_endpoint_and_cert_source(
                options
            )
            assert api_endpoint == mock_api_endpoint
            assert cert_source is None

    # Test cases for mTLS enablement when GOOGLE_API_USE_CLIENT_CERTIFICATE is unset.
    test_cases = [
        (
            # With workloads present in config, mTLS is enabled.
            {
                "version": 1,
                "cert_configs": {
                    "workload": {
                        "cert_path": "path/to/cert/file",
                        "key_path": "path/to/key/file",
                    }
                },
            },
            mock_client_cert_source,
        ),
        (
            # With workloads not present in config, mTLS is disabled.
            {
                "version": 1,
                "cert_configs": {},
            },
            None,
        ),
    ]
    if hasattr(google.auth.transport.mtls, "should_use_client_cert"):
        for config_data, expected_cert_source in test_cases:
            env = os.environ.copy()
            env.pop("GOOGLE_API_USE_CLIENT_CERTIFICATE", None)
            with mock.patch.dict(os.environ, env, clear=True):
                config_filename = "mock_certificate_config.json"
                config_file_content = json.dumps(config_data)
                m = mock.mock_open(read_data=config_file_content)
                with mock.patch("builtins.open", m):
                    with mock.patch.dict(
                        os.environ, {"GOOGLE_API_CERTIFICATE_CONFIG": config_filename}
                    ):
                        mock_api_endpoint = "foo"
                        options = client_options.ClientOptions(
                            client_cert_source=mock_client_cert_source,
                            api_endpoint=mock_api_endpoint,
                        )
                        (
                            api_endpoint,
                            cert_source,
                        ) = client_class.get_mtls_endpoint_and_cert_source(options)
                        assert api_endpoint == mock_api_endpoint
                        assert cert_source is expected_cert_source

    # Test cases for mTLS enablement when GOOGLE_API_USE_CLIENT_CERTIFICATE is unset(empty).
    test_cases = [
        (
            # With workloads present in config, mTLS is enabled.
            {
                "version": 1,
                "cert_configs": {
                    "workload": {
                        "cert_path": "path/to/cert/file",
                        "key_path": "path/to/key/file",
                    }
                },
            },
            mock_client_cert_source,
        ),
        (
            # With workloads not present in config, mTLS is disabled.
            {
                "version": 1,
                "cert_configs": {},
            },
            None,
        ),
    ]
    if hasattr(google.auth.transport.mtls, "should_use_client_cert"):
        for config_data, expected_cert_source in test_cases:
            env = os.environ.copy()
            env.pop("GOOGLE_API_USE_CLIENT_CERTIFICATE", "")
            with mock.patch.dict(os.environ, env, clear=True):
                config_filename = "mock_certificate_config.json"
                config_file_content = json.dumps(config_data)
                m = mock.mock_open(read_data=config_file_content)
                with mock.patch("builtins.open", m):
                    with mock.patch.dict(
                        os.environ, {"GOOGLE_API_CERTIFICATE_CONFIG": config_filename}
                    ):
                        mock_api_endpoint = "foo"
                        options = client_options.ClientOptions(
                            client_cert_source=mock_client_cert_source,
                            api_endpoint=mock_api_endpoint,
                        )
                        (
                            api_endpoint,
                            cert_source,
                        ) = client_class.get_mtls_endpoint_and_cert_source(options)
                        assert api_endpoint == mock_api_endpoint
                        assert cert_source is expected_cert_source

    # Test the case GOOGLE_API_USE_MTLS_ENDPOINT is "never".
    with mock.patch.dict(os.environ, {"GOOGLE_API_USE_MTLS_ENDPOINT": "never"}):
        api_endpoint, cert_source = client_class.get_mtls_endpoint_and_cert_source()
        assert api_endpoint == client_class.DEFAULT_ENDPOINT
        assert cert_source is None

    # Test the case GOOGLE_API_USE_MTLS_ENDPOINT is "always".
    with mock.patch.dict(os.environ, {"GOOGLE_API_USE_MTLS_ENDPOINT": "always"}):
        api_endpoint, cert_source = client_class.get_mtls_endpoint_and_cert_source()
        assert api_endpoint == client_class.DEFAULT_MTLS_ENDPOINT
        assert cert_source is None

    # Test the case GOOGLE_API_USE_MTLS_ENDPOINT is "auto" and default cert doesn't exist.
    with mock.patch.dict(os.environ, {"GOOGLE_API_USE_CLIENT_CERTIFICATE": "true"}):
        with mock.patch(
            "google.auth.transport.mtls.has_default_client_cert_source",
            return_value=False,
        ):
            api_endpoint, cert_source = client_class.get_mtls_endpoint_and_cert_source()
            assert api_endpoint == client_class.DEFAULT_ENDPOINT
            assert cert_source is None

    # Test the case GOOGLE_API_USE_MTLS_ENDPOINT is "auto" and default cert exists.
    with mock.patch.dict(os.environ, {"GOOGLE_API_USE_CLIENT_CERTIFICATE": "true"}):
        with mock.patch(
            "google.auth.transport.mtls.has_default_client_cert_source",
            return_value=True,
        ):
            with mock.patch(
                "google.auth.transport.mtls.default_client_cert_source",
                return_value=mock_client_cert_source,
            ):
                (
                    api_endpoint,
                    cert_source,
                ) = client_class.get_mtls_endpoint_and_cert_source()
                assert api_endpoint == client_class.DEFAULT_MTLS_ENDPOINT
                assert cert_source == mock_client_cert_source

    # Check the case api_endpoint is not provided and GOOGLE_API_USE_MTLS_ENDPOINT has
    # unsupported value.
    with mock.patch.dict(os.environ, {"GOOGLE_API_USE_MTLS_ENDPOINT": "Unsupported"}):
        with pytest.raises(MutualTLSChannelError) as excinfo:
            client_class.get_mtls_endpoint_and_cert_source()

        assert (
            str(excinfo.value)
            == "Environment variable `GOOGLE_API_USE_MTLS_ENDPOINT` must be `never`, `auto` or `always`"
        )


@pytest.mark.parametrize("client_class", [QueryServiceClient, QueryServiceAsyncClient])
@mock.patch.object(
    QueryServiceClient,
    "_DEFAULT_ENDPOINT_TEMPLATE",
    modify_default_endpoint_template(QueryServiceClient),
)
@mock.patch.object(
    QueryServiceAsyncClient,
    "_DEFAULT_ENDPOINT_TEMPLATE",
    modify_default_endpoint_template(QueryServiceAsyncClient),
)
def test_query_service_client_client_api_endpoint(client_class):
    mock_client_cert_source = client_cert_source_callback
    api_override = "foo.com"
    default_universe = QueryServiceClient._DEFAULT_UNIVERSE
    default_endpoint = QueryServiceClient._DEFAULT_ENDPOINT_TEMPLATE.format(
        UNIVERSE_DOMAIN=default_universe
    )
    mock_universe = "bar.com"
    mock_endpoint = QueryServiceClient._DEFAULT_ENDPOINT_TEMPLATE.format(
        UNIVERSE_DOMAIN=mock_universe
    )

    # If ClientOptions.api_endpoint is set and GOOGLE_API_USE_CLIENT_CERTIFICATE="true",
    # use ClientOptions.api_endpoint as the api endpoint regardless.
    with mock.patch.dict(os.environ, {"GOOGLE_API_USE_CLIENT_CERTIFICATE": "true"}):
        with mock.patch(
            "google.auth.transport.requests.AuthorizedSession.configure_mtls_channel"
        ):
            options = client_options.ClientOptions(
                client_cert_source=mock_client_cert_source, api_endpoint=api_override
            )
            client = client_class(
                client_options=options,
                credentials=ga_credentials.AnonymousCredentials(),
            )
            assert client.api_endpoint == api_override

    # If ClientOptions.api_endpoint is not set and GOOGLE_API_USE_MTLS_ENDPOINT="never",
    # use the _DEFAULT_ENDPOINT_TEMPLATE populated with GDU as the api endpoint.
    with mock.patch.dict(os.environ, {"GOOGLE_API_USE_MTLS_ENDPOINT": "never"}):
        client = client_class(credentials=ga_credentials.AnonymousCredentials())
        assert client.api_endpoint == default_endpoint

    # If ClientOptions.api_endpoint is not set and GOOGLE_API_USE_MTLS_ENDPOINT="always",
    # use the DEFAULT_MTLS_ENDPOINT as the api endpoint.
    with mock.patch.dict(os.environ, {"GOOGLE_API_USE_MTLS_ENDPOINT": "always"}):
        client = client_class(credentials=ga_credentials.AnonymousCredentials())
        assert client.api_endpoint == client_class.DEFAULT_MTLS_ENDPOINT

    # If ClientOptions.api_endpoint is not set, GOOGLE_API_USE_MTLS_ENDPOINT="auto" (default),
    # GOOGLE_API_USE_CLIENT_CERTIFICATE="false" (default), default cert source doesn't exist,
    # and ClientOptions.universe_domain="bar.com",
    # use the _DEFAULT_ENDPOINT_TEMPLATE populated with universe domain as the api endpoint.
    options = client_options.ClientOptions()
    universe_exists = hasattr(options, "universe_domain")
    if universe_exists:
        options = client_options.ClientOptions(universe_domain=mock_universe)
        client = client_class(
            client_options=options, credentials=ga_credentials.AnonymousCredentials()
        )
    else:
        client = client_class(
            client_options=options, credentials=ga_credentials.AnonymousCredentials()
        )
    assert client.api_endpoint == (
        mock_endpoint if universe_exists else default_endpoint
    )
    assert client.universe_domain == (
        mock_universe if universe_exists else default_universe
    )

    # If ClientOptions does not have a universe domain attribute and GOOGLE_API_USE_MTLS_ENDPOINT="never",
    # use the _DEFAULT_ENDPOINT_TEMPLATE populated with GDU as the api endpoint.
    options = client_options.ClientOptions()
    if hasattr(options, "universe_domain"):
        delattr(options, "universe_domain")
    with mock.patch.dict(os.environ, {"GOOGLE_API_USE_MTLS_ENDPOINT": "never"}):
        client = client_class(
            client_options=options, credentials=ga_credentials.AnonymousCredentials()
        )
        assert client.api_endpoint == default_endpoint


@pytest.mark.parametrize(
    "client_class,transport_class,transport_name",
    [
        (QueryServiceClient, transports.QueryServiceGrpcTransport, "grpc"),
        (
            QueryServiceAsyncClient,
            transports.QueryServiceGrpcAsyncIOTransport,
            "grpc_asyncio",
        ),
    ],
)
def test_query_service_client_client_options_scopes(
    client_class, transport_class, transport_name
):
    # Check the case scopes are provided.
    options = client_options.ClientOptions(
        scopes=["1", "2"],
    )
    with mock.patch.object(transport_class, "__init__") as patched:
        patched.return_value = None
        client = client_class(client_options=options, transport=transport_name)
        patched.assert_called_once_with(
            credentials=None,
            credentials_file=None,
            host=client._DEFAULT_ENDPOINT_TEMPLATE.format(
                UNIVERSE_DOMAIN=client._DEFAULT_UNIVERSE
            ),
            scopes=["1", "2"],
            client_cert_source_for_mtls=None,
            quota_project_id=None,
            client_info=transports.base.DEFAULT_CLIENT_INFO,
            always_use_jwt_access=True,
            api_audience=None,
        )


@pytest.mark.parametrize(
    "client_class,transport_class,transport_name,grpc_helpers",
    [
        (
            QueryServiceClient,
            transports.QueryServiceGrpcTransport,
            "grpc",
            grpc_helpers,
        ),
        (
            QueryServiceAsyncClient,
            transports.QueryServiceGrpcAsyncIOTransport,
            "grpc_asyncio",
            grpc_helpers_async,
        ),
    ],
)
def test_query_service_client_client_options_credentials_file(
    client_class, transport_class, transport_name, grpc_helpers
):
    # Check the case credentials file is provided.
    options = client_options.ClientOptions(credentials_file="credentials.json")

    with mock.patch.object(transport_class, "__init__") as patched:
        patched.return_value = None
        client = client_class(client_options=options, transport=transport_name)
        patched.assert_called_once_with(
            credentials=None,
            credentials_file="credentials.json",
            host=client._DEFAULT_ENDPOINT_TEMPLATE.format(
                UNIVERSE_DOMAIN=client._DEFAULT_UNIVERSE
            ),
            scopes=None,
            client_cert_source_for_mtls=None,
            quota_project_id=None,
            client_info=transports.base.DEFAULT_CLIENT_INFO,
            always_use_jwt_access=True,
            api_audience=None,
        )


def test_query_service_client_client_options_from_dict():
    with mock.patch(
        "google.cloud.monitoring_v3.services.query_service.transports.QueryServiceGrpcTransport.__init__"
    ) as grpc_transport:
        grpc_transport.return_value = None
        client = QueryServiceClient(client_options={"api_endpoint": "squid.clam.whelk"})
        grpc_transport.assert_called_once_with(
            credentials=None,
            credentials_file=None,
            host="squid.clam.whelk",
            scopes=None,
            client_cert_source_for_mtls=None,
            quota_project_id=None,
            client_info=transports.base.DEFAULT_CLIENT_INFO,
            always_use_jwt_access=True,
            api_audience=None,
        )


@pytest.mark.parametrize(
    "client_class,transport_class,transport_name,grpc_helpers",
    [
        (
            QueryServiceClient,
            transports.QueryServiceGrpcTransport,
            "grpc",
            grpc_helpers,
        ),
        (
            QueryServiceAsyncClient,
            transports.QueryServiceGrpcAsyncIOTransport,
            "grpc_asyncio",
            grpc_helpers_async,
        ),
    ],
)
def test_query_service_client_create_channel_credentials_file(
    client_class, transport_class, transport_name, grpc_helpers
):
    # Check the case credentials file is provided.
    options = client_options.ClientOptions(credentials_file="credentials.json")

    with mock.patch.object(transport_class, "__init__") as patched:
        patched.return_value = None
        client = client_class(client_options=options, transport=transport_name)
        patched.assert_called_once_with(
            credentials=None,
            credentials_file="credentials.json",
            host=client._DEFAULT_ENDPOINT_TEMPLATE.format(
                UNIVERSE_DOMAIN=client._DEFAULT_UNIVERSE
            ),
            scopes=None,
            client_cert_source_for_mtls=None,
            quota_project_id=None,
            client_info=transports.base.DEFAULT_CLIENT_INFO,
            always_use_jwt_access=True,
            api_audience=None,
        )

    # test that the credentials from file are saved and used as the credentials.
    with mock.patch.object(
        google.auth, "load_credentials_from_file", autospec=True
    ) as load_creds, mock.patch.object(
        google.auth, "default", autospec=True
    ) as adc, mock.patch.object(
        grpc_helpers, "create_channel"
    ) as create_channel:
        creds = ga_credentials.AnonymousCredentials()
        file_creds = ga_credentials.AnonymousCredentials()
        load_creds.return_value = (file_creds, None)
        adc.return_value = (creds, None)
        client = client_class(client_options=options, transport=transport_name)
        create_channel.assert_called_with(
            "monitoring.googleapis.com:443",
            credentials=file_creds,
            credentials_file=None,
            quota_project_id=None,
            default_scopes=(
                "https://www.googleapis.com/auth/cloud-platform",
                "https://www.googleapis.com/auth/monitoring",
                "https://www.googleapis.com/auth/monitoring.read",
            ),
            scopes=None,
            default_host="monitoring.googleapis.com",
            ssl_credentials=None,
            options=[
                ("grpc.max_send_message_length", -1),
                ("grpc.max_receive_message_length", -1),
            ],
        )


@pytest.mark.parametrize(
    "request_type",
    [
        metric_service.QueryTimeSeriesRequest,
        dict,
    ],
)
def test_query_time_series(request_type, transport: str = "grpc"):
    client = QueryServiceClient(
        credentials=ga_credentials.AnonymousCredentials(),
        transport=transport,
    )

    # Everything is optional in proto3 as far as the runtime is concerned,
    # and we are mocking out the actual API, so just send an empty request.
    request = request_type()

    # Mock the actual call within the gRPC stub, and fake the request.
    with mock.patch.object(
        type(client.transport.query_time_series), "__call__"
    ) as call:
        # Designate an appropriate return value for the call.
        call.return_value = metric_service.QueryTimeSeriesResponse(
            next_page_token="next_page_token_value",
        )
        response = client.query_time_series(request)

        # Establish that the underlying gRPC stub method was called.
        assert len(call.mock_calls) == 1
        _, args, _ = call.mock_calls[0]
        request = metric_service.QueryTimeSeriesRequest()
        assert args[0] == request

    # Establish that the response is the type that we expect.
    assert isinstance(response, pagers.QueryTimeSeriesPager)
    assert response.next_page_token == "next_page_token_value"


def test_query_time_series_non_empty_request_with_auto_populated_field():
    # This test is a coverage failsafe to make sure that UUID4 fields are
    # automatically populated, according to AIP-4235, with non-empty requests.
    client = QueryServiceClient(
        credentials=ga_credentials.AnonymousCredentials(),
        transport="grpc",
    )

    # Populate all string fields in the request which are not UUID4
    # since we want to check that UUID4 are populated automatically
    # if they meet the requirements of AIP 4235.
    request = metric_service.QueryTimeSeriesRequest(
        name="name_value",
        query="query_value",
        page_token="page_token_value",
    )

    # Mock the actual call within the gRPC stub, and fake the request.
    with mock.patch.object(
        type(client.transport.query_time_series), "__call__"
    ) as call:
        call.return_value.name = (
            "foo"  # operation_request.operation in compute client(s) expect a string.
        )
        client.query_time_series(request=request)
        call.assert_called()
        _, args, _ = call.mock_calls[0]
        assert args[0] == metric_service.QueryTimeSeriesRequest(
            name="name_value",
            query="query_value",
            page_token="page_token_value",
        )


def test_query_time_series_use_cached_wrapped_rpc():
    # Clients should use _prep_wrapped_messages to create cached wrapped rpcs,
    # instead of constructing them on each call
    with mock.patch("google.api_core.gapic_v1.method.wrap_method") as wrapper_fn:
        client = QueryServiceClient(
            credentials=ga_credentials.AnonymousCredentials(),
            transport="grpc",
        )

        # Should wrap all calls on client creation
        assert wrapper_fn.call_count > 0
        wrapper_fn.reset_mock()

        # Ensure method has been cached
        assert client._transport.query_time_series in client._transport._wrapped_methods

        # Replace cached wrapped function with mock
        mock_rpc = mock.Mock()
        mock_rpc.return_value.name = (
            "foo"  # operation_request.operation in compute client(s) expect a string.
        )
        client._transport._wrapped_methods[
            client._transport.query_time_series
        ] = mock_rpc
        request = {}
        client.query_time_series(request)

        # Establish that the underlying gRPC stub method was called.
        assert mock_rpc.call_count == 1

        client.query_time_series(request)

        # Establish that a new wrapper was not created for this call
        assert wrapper_fn.call_count == 0
        assert mock_rpc.call_count == 2


@pytest.mark.asyncio
async def test_query_time_series_async_use_cached_wrapped_rpc(
    transport: str = "grpc_asyncio",
):
    # Clients should use _prep_wrapped_messages to create cached wrapped rpcs,
    # instead of constructing them on each call
    with mock.patch("google.api_core.gapic_v1.method_async.wrap_method") as wrapper_fn:
        client = QueryServiceAsyncClient(
            credentials=async_anonymous_credentials(),
            transport=transport,
        )

        # Should wrap all calls on client creation
        assert wrapper_fn.call_count > 0
        wrapper_fn.reset_mock()

        # Ensure method has been cached
        assert (
            client._client._transport.query_time_series
            in client._client._transport._wrapped_methods
        )

        # Replace cached wrapped function with mock
        mock_rpc = mock.AsyncMock()
        mock_rpc.return_value = mock.Mock()
        client._client._transport._wrapped_methods[
            client._client._transport.query_time_series
        ] = mock_rpc

        request = {}
        await client.query_time_series(request)

        # Establish that the underlying gRPC stub method was called.
        assert mock_rpc.call_count == 1

        await client.query_time_series(request)

        # Establish that a new wrapper was not created for this call
        assert wrapper_fn.call_count == 0
        assert mock_rpc.call_count == 2


@pytest.mark.asyncio
async def test_query_time_series_async(
    transport: str = "grpc_asyncio", request_type=metric_service.QueryTimeSeriesRequest
):
    client = QueryServiceAsyncClient(
        credentials=async_anonymous_credentials(),
        transport=transport,
    )

    # Everything is optional in proto3 as far as the runtime is concerned,
    # and we are mocking out the actual API, so just send an empty request.
    request = request_type()

    # Mock the actual call within the gRPC stub, and fake the request.
    with mock.patch.object(
        type(client.transport.query_time_series), "__call__"
    ) as call:
        # Designate an appropriate return value for the call.
        call.return_value = grpc_helpers_async.FakeUnaryUnaryCall(
            metric_service.QueryTimeSeriesResponse(
                next_page_token="next_page_token_value",
            )
        )
        response = await client.query_time_series(request)

        # Establish that the underlying gRPC stub method was called.
        assert len(call.mock_calls)
        _, args, _ = call.mock_calls[0]
        request = metric_service.QueryTimeSeriesRequest()
        assert args[0] == request

    # Establish that the response is the type that we expect.
    assert isinstance(response, pagers.QueryTimeSeriesAsyncPager)
    assert response.next_page_token == "next_page_token_value"


@pytest.mark.asyncio
async def test_query_time_series_async_from_dict():
    await test_query_time_series_async(request_type=dict)


def test_query_time_series_field_headers():
    client = QueryServiceClient(
        credentials=ga_credentials.AnonymousCredentials(),
    )

    # Any value that is part of the HTTP/1.1 URI should be sent as
    # a field header. Set these to a non-empty value.
    request = metric_service.QueryTimeSeriesRequest()

    request.name = "name_value"

    # Mock the actual call within the gRPC stub, and fake the request.
    with mock.patch.object(
        type(client.transport.query_time_series), "__call__"
    ) as call:
        call.return_value = metric_service.QueryTimeSeriesResponse()
        client.query_time_series(request)

        # Establish that the underlying gRPC stub method was called.
        assert len(call.mock_calls) == 1
        _, args, _ = call.mock_calls[0]
        assert args[0] == request

    # Establish that the field header was sent.
    _, _, kw = call.mock_calls[0]
    assert (
        "x-goog-request-params",
        "name=name_value",
    ) in kw["metadata"]


@pytest.mark.asyncio
async def test_query_time_series_field_headers_async():
    client = QueryServiceAsyncClient(
        credentials=async_anonymous_credentials(),
    )

    # Any value that is part of the HTTP/1.1 URI should be sent as
    # a field header. Set these to a non-empty value.
    request = metric_service.QueryTimeSeriesRequest()

    request.name = "name_value"

    # Mock the actual call within the gRPC stub, and fake the request.
    with mock.patch.object(
        type(client.transport.query_time_series), "__call__"
    ) as call:
        call.return_value = grpc_helpers_async.FakeUnaryUnaryCall(
            metric_service.QueryTimeSeriesResponse()
        )
        await client.query_time_series(request)

        # Establish that the underlying gRPC stub method was called.
        assert len(call.mock_calls)
        _, args, _ = call.mock_calls[0]
        assert args[0] == request

    # Establish that the field header was sent.
    _, _, kw = call.mock_calls[0]
    assert (
        "x-goog-request-params",
        "name=name_value",
    ) in kw["metadata"]


def test_query_time_series_pager(transport_name: str = "grpc"):
    client = QueryServiceClient(
        credentials=ga_credentials.AnonymousCredentials(),
        transport=transport_name,
    )

    # Mock the actual call within the gRPC stub, and fake the request.
    with mock.patch.object(
        type(client.transport.query_time_series), "__call__"
    ) as call:
        # Set the response to a series of pages.
        call.side_effect = (
            metric_service.QueryTimeSeriesResponse(
                time_series_data=[
                    metric.TimeSeriesData(),
                    metric.TimeSeriesData(),
                    metric.TimeSeriesData(),
                ],
                next_page_token="abc",
            ),
            metric_service.QueryTimeSeriesResponse(
                time_series_data=[],
                next_page_token="def",
            ),
            metric_service.QueryTimeSeriesResponse(
                time_series_data=[
                    metric.TimeSeriesData(),
                ],
                next_page_token="ghi",
            ),
            metric_service.QueryTimeSeriesResponse(
                time_series_data=[
                    metric.TimeSeriesData(),
                    metric.TimeSeriesData(),
                ],
            ),
            RuntimeError,
        )

        expected_metadata = ()
        retry = retries.Retry()
        timeout = 5
        expected_metadata = tuple(expected_metadata) + (
            gapic_v1.routing_header.to_grpc_metadata((("name", ""),)),
        )
        pager = client.query_time_series(request={}, retry=retry, timeout=timeout)

        assert pager._metadata == expected_metadata
        assert pager._retry == retry
        assert pager._timeout == timeout

        results = list(pager)
        assert len(results) == 6
        assert all(isinstance(i, metric.TimeSeriesData) for i in results)


def test_query_time_series_pages(transport_name: str = "grpc"):
    client = QueryServiceClient(
        credentials=ga_credentials.AnonymousCredentials(),
        transport=transport_name,
    )

    # Mock the actual call within the gRPC stub, and fake the request.
    with mock.patch.object(
        type(client.transport.query_time_series), "__call__"
    ) as call:
        # Set the response to a series of pages.
        call.side_effect = (
            metric_service.QueryTimeSeriesResponse(
                time_series_data=[
                    metric.TimeSeriesData(),
                    metric.TimeSeriesData(),
                    metric.TimeSeriesData(),
                ],
                next_page_token="abc",
            ),
            metric_service.QueryTimeSeriesResponse(
                time_series_data=[],
                next_page_token="def",
            ),
            metric_service.QueryTimeSeriesResponse(
                time_series_data=[
                    metric.TimeSeriesData(),
                ],
                next_page_token="ghi",
            ),
            metric_service.QueryTimeSeriesResponse(
                time_series_data=[
                    metric.TimeSeriesData(),
                    metric.TimeSeriesData(),
                ],
            ),
            RuntimeError,
        )
        pages = list(client.query_time_series(request={}).pages)
        for page_, token in zip(pages, ["abc", "def", "ghi", ""]):
            assert page_.raw_page.next_page_token == token


@pytest.mark.asyncio
async def test_query_time_series_async_pager():
    client = QueryServiceAsyncClient(
        credentials=async_anonymous_credentials(),
    )

    # Mock the actual call within the gRPC stub, and fake the request.
    with mock.patch.object(
        type(client.transport.query_time_series),
        "__call__",
        new_callable=mock.AsyncMock,
    ) as call:
        # Set the response to a series of pages.
        call.side_effect = (
            metric_service.QueryTimeSeriesResponse(
                time_series_data=[
                    metric.TimeSeriesData(),
                    metric.TimeSeriesData(),
                    metric.TimeSeriesData(),
                ],
                next_page_token="abc",
            ),
            metric_service.QueryTimeSeriesResponse(
                time_series_data=[],
                next_page_token="def",
            ),
            metric_service.QueryTimeSeriesResponse(
                time_series_data=[
                    metric.TimeSeriesData(),
                ],
                next_page_token="ghi",
            ),
            metric_service.QueryTimeSeriesResponse(
                time_series_data=[
                    metric.TimeSeriesData(),
                    metric.TimeSeriesData(),
                ],
            ),
            RuntimeError,
        )
        async_pager = await client.query_time_series(
            request={},
        )
        assert async_pager.next_page_token == "abc"
        responses = []
        async for response in async_pager:  # pragma: no branch
            responses.append(response)

        assert len(responses) == 6
        assert all(isinstance(i, metric.TimeSeriesData) for i in responses)


@pytest.mark.asyncio
async def test_query_time_series_async_pages():
    client = QueryServiceAsyncClient(
        credentials=async_anonymous_credentials(),
    )

    # Mock the actual call within the gRPC stub, and fake the request.
    with mock.patch.object(
        type(client.transport.query_time_series),
        "__call__",
        new_callable=mock.AsyncMock,
    ) as call:
        # Set the response to a series of pages.
        call.side_effect = (
            metric_service.QueryTimeSeriesResponse(
                time_series_data=[
                    metric.TimeSeriesData(),
                    metric.TimeSeriesData(),
                    metric.TimeSeriesData(),
                ],
                next_page_token="abc",
            ),
            metric_service.QueryTimeSeriesResponse(
                time_series_data=[],
                next_page_token="def",
            ),
            metric_service.QueryTimeSeriesResponse(
                time_series_data=[
                    metric.TimeSeriesData(),
                ],
                next_page_token="ghi",
            ),
            metric_service.QueryTimeSeriesResponse(
                time_series_data=[
                    metric.TimeSeriesData(),
                    metric.TimeSeriesData(),
                ],
            ),
            RuntimeError,
        )
        pages = []
        # Workaround issue in python 3.9 related to code coverage by adding `# pragma: no branch`
        # See https://github.com/googleapis/gapic-generator-python/pull/1174#issuecomment-1025132372
        async for page_ in (  # pragma: no branch
            await client.query_time_series(request={})
        ).pages:
            pages.append(page_)
        for page_, token in zip(pages, ["abc", "def", "ghi", ""]):
            assert page_.raw_page.next_page_token == token


def test_credentials_transport_error():
    # It is an error to provide credentials and a transport instance.
    transport = transports.QueryServiceGrpcTransport(
        credentials=ga_credentials.AnonymousCredentials(),
    )
    with pytest.raises(ValueError):
        client = QueryServiceClient(
            credentials=ga_credentials.AnonymousCredentials(),
            transport=transport,
        )

    # It is an error to provide a credentials file and a transport instance.
    transport = transports.QueryServiceGrpcTransport(
        credentials=ga_credentials.AnonymousCredentials(),
    )
    with pytest.raises(ValueError):
        client = QueryServiceClient(
            client_options={"credentials_file": "credentials.json"},
            transport=transport,
        )

    # It is an error to provide an api_key and a transport instance.
    transport = transports.QueryServiceGrpcTransport(
        credentials=ga_credentials.AnonymousCredentials(),
    )
    options = client_options.ClientOptions()
    options.api_key = "api_key"
    with pytest.raises(ValueError):
        client = QueryServiceClient(
            client_options=options,
            transport=transport,
        )

    # It is an error to provide an api_key and a credential.
    options = client_options.ClientOptions()
    options.api_key = "api_key"
    with pytest.raises(ValueError):
        client = QueryServiceClient(
            client_options=options, credentials=ga_credentials.AnonymousCredentials()
        )

    # It is an error to provide scopes and a transport instance.
    transport = transports.QueryServiceGrpcTransport(
        credentials=ga_credentials.AnonymousCredentials(),
    )
    with pytest.raises(ValueError):
        client = QueryServiceClient(
            client_options={"scopes": ["1", "2"]},
            transport=transport,
        )


def test_transport_instance():
    # A client may be instantiated with a custom transport instance.
    transport = transports.QueryServiceGrpcTransport(
        credentials=ga_credentials.AnonymousCredentials(),
    )
    client = QueryServiceClient(transport=transport)
    assert client.transport is transport


def test_transport_get_channel():
    # A client may be instantiated with a custom transport instance.
    transport = transports.QueryServiceGrpcTransport(
        credentials=ga_credentials.AnonymousCredentials(),
    )
    channel = transport.grpc_channel
    assert channel

    transport = transports.QueryServiceGrpcAsyncIOTransport(
        credentials=ga_credentials.AnonymousCredentials(),
    )
    channel = transport.grpc_channel
    assert channel


@pytest.mark.parametrize(
    "transport_class",
    [
        transports.QueryServiceGrpcTransport,
        transports.QueryServiceGrpcAsyncIOTransport,
    ],
)
def test_transport_adc(transport_class):
    # Test default credentials are used if not provided.
    with mock.patch.object(google.auth, "default") as adc:
        adc.return_value = (ga_credentials.AnonymousCredentials(), None)
        transport_class()
        adc.assert_called_once()


def test_transport_kind_grpc():
    transport = QueryServiceClient.get_transport_class("grpc")(
        credentials=ga_credentials.AnonymousCredentials()
    )
    assert transport.kind == "grpc"


def test_initialize_client_w_grpc():
    client = QueryServiceClient(
        credentials=ga_credentials.AnonymousCredentials(), transport="grpc"
    )
    assert client is not None


# This test is a coverage failsafe to make sure that totally empty calls,
# i.e. request == None and no flattened fields passed, work.
def test_query_time_series_empty_call_grpc():
    client = QueryServiceClient(
        credentials=ga_credentials.AnonymousCredentials(),
        transport="grpc",
    )

    # Mock the actual call, and fake the request.
    with mock.patch.object(
        type(client.transport.query_time_series), "__call__"
    ) as call:
        call.return_value = metric_service.QueryTimeSeriesResponse()
        client.query_time_series(request=None)

        # Establish that the underlying stub method was called.
        call.assert_called()
        _, args, _ = call.mock_calls[0]
        request_msg = metric_service.QueryTimeSeriesRequest()

        assert args[0] == request_msg


def test_transport_kind_grpc_asyncio():
    transport = QueryServiceAsyncClient.get_transport_class("grpc_asyncio")(
        credentials=async_anonymous_credentials()
    )
    assert transport.kind == "grpc_asyncio"


def test_initialize_client_w_grpc_asyncio():
    client = QueryServiceAsyncClient(
        credentials=async_anonymous_credentials(), transport="grpc_asyncio"
    )
    assert client is not None


# This test is a coverage failsafe to make sure that totally empty calls,
# i.e. request == None and no flattened fields passed, work.
@pytest.mark.asyncio
async def test_query_time_series_empty_call_grpc_asyncio():
    client = QueryServiceAsyncClient(
        credentials=async_anonymous_credentials(),
        transport="grpc_asyncio",
    )

    # Mock the actual call, and fake the request.
    with mock.patch.object(
        type(client.transport.query_time_series), "__call__"
    ) as call:
        # Designate an appropriate return value for the call.
        call.return_value = grpc_helpers_async.FakeUnaryUnaryCall(
            metric_service.QueryTimeSeriesResponse(
                next_page_token="next_page_token_value",
            )
        )
        await client.query_time_series(request=None)

        # Establish that the underlying stub method was called.
        call.assert_called()
        _, args, _ = call.mock_calls[0]
        request_msg = metric_service.QueryTimeSeriesRequest()

        assert args[0] == request_msg


def test_transport_grpc_default():
    # A client should use the gRPC transport by default.
    client = QueryServiceClient(
        credentials=ga_credentials.AnonymousCredentials(),
    )
    assert isinstance(
        client.transport,
        transports.QueryServiceGrpcTransport,
    )


def test_query_service_base_transport_error():
    # Passing both a credentials object and credentials_file should raise an error
    with pytest.raises(core_exceptions.DuplicateCredentialArgs):
        transport = transports.QueryServiceTransport(
            credentials=ga_credentials.AnonymousCredentials(),
            credentials_file="credentials.json",
        )


def test_query_service_base_transport():
    # Instantiate the base transport.
    with mock.patch(
        "google.cloud.monitoring_v3.services.query_service.transports.QueryServiceTransport.__init__"
    ) as Transport:
        Transport.return_value = None
        transport = transports.QueryServiceTransport(
            credentials=ga_credentials.AnonymousCredentials(),
        )

    # Every method on the transport should just blindly
    # raise NotImplementedError.
    methods = ("query_time_series",)
    for method in methods:
        with pytest.raises(NotImplementedError):
            getattr(transport, method)(request=object())

    with pytest.raises(NotImplementedError):
        transport.close()

    # Catch all for all remaining methods and properties
    remainder = [
        "kind",
    ]
    for r in remainder:
        with pytest.raises(NotImplementedError):
            getattr(transport, r)()


def test_query_service_base_transport_with_credentials_file():
    # Instantiate the base transport with a credentials file
    with mock.patch.object(
        google.auth, "load_credentials_from_file", autospec=True
    ) as load_creds, mock.patch(
        "google.cloud.monitoring_v3.services.query_service.transports.QueryServiceTransport._prep_wrapped_messages"
    ) as Transport:
        Transport.return_value = None
        load_creds.return_value = (ga_credentials.AnonymousCredentials(), None)
        transport = transports.QueryServiceTransport(
            credentials_file="credentials.json",
            quota_project_id="octopus",
        )
        load_creds.assert_called_once_with(
            "credentials.json",
            scopes=None,
            default_scopes=(
                "https://www.googleapis.com/auth/cloud-platform",
                "https://www.googleapis.com/auth/monitoring",
                "https://www.googleapis.com/auth/monitoring.read",
            ),
            quota_project_id="octopus",
        )


def test_query_service_base_transport_with_adc():
    # Test the default credentials are used if credentials and credentials_file are None.
    with mock.patch.object(google.auth, "default", autospec=True) as adc, mock.patch(
        "google.cloud.monitoring_v3.services.query_service.transports.QueryServiceTransport._prep_wrapped_messages"
    ) as Transport:
        Transport.return_value = None
        adc.return_value = (ga_credentials.AnonymousCredentials(), None)
        transport = transports.QueryServiceTransport()
        adc.assert_called_once()


def test_query_service_auth_adc():
    # If no credentials are provided, we should use ADC credentials.
    with mock.patch.object(google.auth, "default", autospec=True) as adc:
        adc.return_value = (ga_credentials.AnonymousCredentials(), None)
        QueryServiceClient()
        adc.assert_called_once_with(
            scopes=None,
            default_scopes=(
                "https://www.googleapis.com/auth/cloud-platform",
                "https://www.googleapis.com/auth/monitoring",
                "https://www.googleapis.com/auth/monitoring.read",
            ),
            quota_project_id=None,
        )


@pytest.mark.parametrize(
    "transport_class",
    [
        transports.QueryServiceGrpcTransport,
        transports.QueryServiceGrpcAsyncIOTransport,
    ],
)
def test_query_service_transport_auth_adc(transport_class):
    # If credentials and host are not provided, the transport class should use
    # ADC credentials.
    with mock.patch.object(google.auth, "default", autospec=True) as adc:
        adc.return_value = (ga_credentials.AnonymousCredentials(), None)
        transport_class(quota_project_id="octopus", scopes=["1", "2"])
        adc.assert_called_once_with(
            scopes=["1", "2"],
            default_scopes=(
                "https://www.googleapis.com/auth/cloud-platform",
                "https://www.googleapis.com/auth/monitoring",
                "https://www.googleapis.com/auth/monitoring.read",
            ),
            quota_project_id="octopus",
        )


@pytest.mark.parametrize(
    "transport_class",
    [
        transports.QueryServiceGrpcTransport,
        transports.QueryServiceGrpcAsyncIOTransport,
    ],
)
def test_query_service_transport_auth_gdch_credentials(transport_class):
    host = "https://language.com"
    api_audience_tests = [None, "https://language2.com"]
    api_audience_expect = [host, "https://language2.com"]
    for t, e in zip(api_audience_tests, api_audience_expect):
        with mock.patch.object(google.auth, "default", autospec=True) as adc:
            gdch_mock = mock.MagicMock()
            type(gdch_mock).with_gdch_audience = mock.PropertyMock(
                return_value=gdch_mock
            )
            adc.return_value = (gdch_mock, None)
            transport_class(host=host, api_audience=t)
            gdch_mock.with_gdch_audience.assert_called_once_with(e)


@pytest.mark.parametrize(
    "transport_class,grpc_helpers",
    [
        (transports.QueryServiceGrpcTransport, grpc_helpers),
        (transports.QueryServiceGrpcAsyncIOTransport, grpc_helpers_async),
    ],
)
def test_query_service_transport_create_channel(transport_class, grpc_helpers):
    # If credentials and host are not provided, the transport class should use
    # ADC credentials.
    with mock.patch.object(
        google.auth, "default", autospec=True
    ) as adc, mock.patch.object(
        grpc_helpers, "create_channel", autospec=True
    ) as create_channel:
        creds = ga_credentials.AnonymousCredentials()
        adc.return_value = (creds, None)
        transport_class(quota_project_id="octopus", scopes=["1", "2"])

        create_channel.assert_called_with(
            "monitoring.googleapis.com:443",
            credentials=creds,
            credentials_file=None,
            quota_project_id="octopus",
            default_scopes=(
                "https://www.googleapis.com/auth/cloud-platform",
                "https://www.googleapis.com/auth/monitoring",
                "https://www.googleapis.com/auth/monitoring.read",
            ),
            scopes=["1", "2"],
            default_host="monitoring.googleapis.com",
            ssl_credentials=None,
            options=[
                ("grpc.max_send_message_length", -1),
                ("grpc.max_receive_message_length", -1),
            ],
        )


@pytest.mark.parametrize(
    "transport_class",
    [transports.QueryServiceGrpcTransport, transports.QueryServiceGrpcAsyncIOTransport],
)
def test_query_service_grpc_transport_client_cert_source_for_mtls(transport_class):
    cred = ga_credentials.AnonymousCredentials()

    # Check ssl_channel_credentials is used if provided.
    with mock.patch.object(transport_class, "create_channel") as mock_create_channel:
        mock_ssl_channel_creds = mock.Mock()
        transport_class(
            host="squid.clam.whelk",
            credentials=cred,
            ssl_channel_credentials=mock_ssl_channel_creds,
        )
        mock_create_channel.assert_called_once_with(
            "squid.clam.whelk:443",
            credentials=cred,
            credentials_file=None,
            scopes=None,
            ssl_credentials=mock_ssl_channel_creds,
            quota_project_id=None,
            options=[
                ("grpc.max_send_message_length", -1),
                ("grpc.max_receive_message_length", -1),
            ],
        )

    # Check if ssl_channel_credentials is not provided, then client_cert_source_for_mtls
    # is used.
    with mock.patch.object(transport_class, "create_channel", return_value=mock.Mock()):
        with mock.patch("grpc.ssl_channel_credentials") as mock_ssl_cred:
            transport_class(
                credentials=cred,
                client_cert_source_for_mtls=client_cert_source_callback,
            )
            expected_cert, expected_key = client_cert_source_callback()
            mock_ssl_cred.assert_called_once_with(
                certificate_chain=expected_cert, private_key=expected_key
            )


@pytest.mark.parametrize(
    "transport_name",
    [
        "grpc",
        "grpc_asyncio",
    ],
)
def test_query_service_host_no_port(transport_name):
    client = QueryServiceClient(
        credentials=ga_credentials.AnonymousCredentials(),
        client_options=client_options.ClientOptions(
            api_endpoint="monitoring.googleapis.com"
        ),
        transport=transport_name,
    )
    assert client.transport._host == ("monitoring.googleapis.com:443")


@pytest.mark.parametrize(
    "transport_name",
    [
        "grpc",
        "grpc_asyncio",
    ],
)
def test_query_service_host_with_port(transport_name):
    client = QueryServiceClient(
        credentials=ga_credentials.AnonymousCredentials(),
        client_options=client_options.ClientOptions(
            api_endpoint="monitoring.googleapis.com:8000"
        ),
        transport=transport_name,
    )
    assert client.transport._host == ("monitoring.googleapis.com:8000")


def test_query_service_grpc_transport_channel():
    channel = grpc.secure_channel("http://localhost/", grpc.local_channel_credentials())

    # Check that channel is used if provided.
    transport = transports.QueryServiceGrpcTransport(
        host="squid.clam.whelk",
        channel=channel,
    )
    assert transport.grpc_channel == channel
    assert transport._host == "squid.clam.whelk:443"
    assert transport._ssl_channel_credentials == None


def test_query_service_grpc_asyncio_transport_channel():
    channel = aio.secure_channel("http://localhost/", grpc.local_channel_credentials())

    # Check that channel is used if provided.
    transport = transports.QueryServiceGrpcAsyncIOTransport(
        host="squid.clam.whelk",
        channel=channel,
    )
    assert transport.grpc_channel == channel
    assert transport._host == "squid.clam.whelk:443"
    assert transport._ssl_channel_credentials == None


# Remove this test when deprecated arguments (api_mtls_endpoint, client_cert_source) are
# removed from grpc/grpc_asyncio transport constructor.
@pytest.mark.filterwarnings("ignore::FutureWarning")
@pytest.mark.parametrize(
    "transport_class",
    [transports.QueryServiceGrpcTransport, transports.QueryServiceGrpcAsyncIOTransport],
)
def test_query_service_transport_channel_mtls_with_client_cert_source(transport_class):
    with mock.patch(
        "grpc.ssl_channel_credentials", autospec=True
    ) as grpc_ssl_channel_cred:
        with mock.patch.object(
            transport_class, "create_channel"
        ) as grpc_create_channel:
            mock_ssl_cred = mock.Mock()
            grpc_ssl_channel_cred.return_value = mock_ssl_cred

            mock_grpc_channel = mock.Mock()
            grpc_create_channel.return_value = mock_grpc_channel

            cred = ga_credentials.AnonymousCredentials()
            with pytest.warns(DeprecationWarning):
                with mock.patch.object(google.auth, "default") as adc:
                    adc.return_value = (cred, None)
                    transport = transport_class(
                        host="squid.clam.whelk",
                        api_mtls_endpoint="mtls.squid.clam.whelk",
                        client_cert_source=client_cert_source_callback,
                    )
                    adc.assert_called_once()

            grpc_ssl_channel_cred.assert_called_once_with(
                certificate_chain=b"cert bytes", private_key=b"key bytes"
            )
            grpc_create_channel.assert_called_once_with(
                "mtls.squid.clam.whelk:443",
                credentials=cred,
                credentials_file=None,
                scopes=None,
                ssl_credentials=mock_ssl_cred,
                quota_project_id=None,
                options=[
                    ("grpc.max_send_message_length", -1),
                    ("grpc.max_receive_message_length", -1),
                ],
            )
            assert transport.grpc_channel == mock_grpc_channel
            assert transport._ssl_channel_credentials == mock_ssl_cred


# Remove this test when deprecated arguments (api_mtls_endpoint, client_cert_source) are
# removed from grpc/grpc_asyncio transport constructor.
@pytest.mark.parametrize(
    "transport_class",
    [transports.QueryServiceGrpcTransport, transports.QueryServiceGrpcAsyncIOTransport],
)
def test_query_service_transport_channel_mtls_with_adc(transport_class):
    mock_ssl_cred = mock.Mock()
    with mock.patch.multiple(
        "google.auth.transport.grpc.SslCredentials",
        __init__=mock.Mock(return_value=None),
        ssl_credentials=mock.PropertyMock(return_value=mock_ssl_cred),
    ):
        with mock.patch.object(
            transport_class, "create_channel"
        ) as grpc_create_channel:
            mock_grpc_channel = mock.Mock()
            grpc_create_channel.return_value = mock_grpc_channel
            mock_cred = mock.Mock()

            with pytest.warns(DeprecationWarning):
                transport = transport_class(
                    host="squid.clam.whelk",
                    credentials=mock_cred,
                    api_mtls_endpoint="mtls.squid.clam.whelk",
                    client_cert_source=None,
                )

            grpc_create_channel.assert_called_once_with(
                "mtls.squid.clam.whelk:443",
                credentials=mock_cred,
                credentials_file=None,
                scopes=None,
                ssl_credentials=mock_ssl_cred,
                quota_project_id=None,
                options=[
                    ("grpc.max_send_message_length", -1),
                    ("grpc.max_receive_message_length", -1),
                ],
            )
            assert transport.grpc_channel == mock_grpc_channel


def test_common_billing_account_path():
    billing_account = "squid"
    expected = "billingAccounts/{billing_account}".format(
        billing_account=billing_account,
    )
    actual = QueryServiceClient.common_billing_account_path(billing_account)
    assert expected == actual


def test_parse_common_billing_account_path():
    expected = {
        "billing_account": "clam",
    }
    path = QueryServiceClient.common_billing_account_path(**expected)

    # Check that the path construction is reversible.
    actual = QueryServiceClient.parse_common_billing_account_path(path)
    assert expected == actual


def test_common_folder_path():
    folder = "whelk"
    expected = "folders/{folder}".format(
        folder=folder,
    )
    actual = QueryServiceClient.common_folder_path(folder)
    assert expected == actual


def test_parse_common_folder_path():
    expected = {
        "folder": "octopus",
    }
    path = QueryServiceClient.common_folder_path(**expected)

    # Check that the path construction is reversible.
    actual = QueryServiceClient.parse_common_folder_path(path)
    assert expected == actual


def test_common_organization_path():
    organization = "oyster"
    expected = "organizations/{organization}".format(
        organization=organization,
    )
    actual = QueryServiceClient.common_organization_path(organization)
    assert expected == actual


def test_parse_common_organization_path():
    expected = {
        "organization": "nudibranch",
    }
    path = QueryServiceClient.common_organization_path(**expected)

    # Check that the path construction is reversible.
    actual = QueryServiceClient.parse_common_organization_path(path)
    assert expected == actual


def test_common_project_path():
    project = "cuttlefish"
    expected = "projects/{project}".format(
        project=project,
    )
    actual = QueryServiceClient.common_project_path(project)
    assert expected == actual


def test_parse_common_project_path():
    expected = {
        "project": "mussel",
    }
    path = QueryServiceClient.common_project_path(**expected)

    # Check that the path construction is reversible.
    actual = QueryServiceClient.parse_common_project_path(path)
    assert expected == actual


def test_common_location_path():
    project = "winkle"
    location = "nautilus"
    expected = "projects/{project}/locations/{location}".format(
        project=project,
        location=location,
    )
    actual = QueryServiceClient.common_location_path(project, location)
    assert expected == actual


def test_parse_common_location_path():
    expected = {
        "project": "scallop",
        "location": "abalone",
    }
    path = QueryServiceClient.common_location_path(**expected)

    # Check that the path construction is reversible.
    actual = QueryServiceClient.parse_common_location_path(path)
    assert expected == actual


def test_client_with_default_client_info():
    client_info = gapic_v1.client_info.ClientInfo()

    with mock.patch.object(
        transports.QueryServiceTransport, "_prep_wrapped_messages"
    ) as prep:
        client = QueryServiceClient(
            credentials=ga_credentials.AnonymousCredentials(),
            client_info=client_info,
        )
        prep.assert_called_once_with(client_info)

    with mock.patch.object(
        transports.QueryServiceTransport, "_prep_wrapped_messages"
    ) as prep:
        transport_class = QueryServiceClient.get_transport_class()
        transport = transport_class(
            credentials=ga_credentials.AnonymousCredentials(),
            client_info=client_info,
        )
        prep.assert_called_once_with(client_info)


def test_transport_close_grpc():
    client = QueryServiceClient(
        credentials=ga_credentials.AnonymousCredentials(), transport="grpc"
    )
    with mock.patch.object(
        type(getattr(client.transport, "_grpc_channel")), "close"
    ) as close:
        with client:
            close.assert_not_called()
        close.assert_called_once()


@pytest.mark.asyncio
async def test_transport_close_grpc_asyncio():
    client = QueryServiceAsyncClient(
        credentials=async_anonymous_credentials(), transport="grpc_asyncio"
    )
    with mock.patch.object(
        type(getattr(client.transport, "_grpc_channel")), "close"
    ) as close:
        async with client:
            close.assert_not_called()
        close.assert_called_once()


def test_client_ctx():
    transports = [
        "grpc",
    ]
    for transport in transports:
        client = QueryServiceClient(
            credentials=ga_credentials.AnonymousCredentials(), transport=transport
        )
        # Test client calls underlying transport.
        with mock.patch.object(type(client.transport), "close") as close:
            close.assert_not_called()
            with client:
                pass
            close.assert_called()


@pytest.mark.parametrize(
    "client_class,transport_class",
    [
        (QueryServiceClient, transports.QueryServiceGrpcTransport),
        (QueryServiceAsyncClient, transports.QueryServiceGrpcAsyncIOTransport),
    ],
)
def test_api_key_credentials(client_class, transport_class):
    with mock.patch.object(
        google.auth._default, "get_api_key_credentials", create=True
    ) as get_api_key_credentials:
        mock_cred = mock.Mock()
        get_api_key_credentials.return_value = mock_cred
        options = client_options.ClientOptions()
        options.api_key = "api_key"
        with mock.patch.object(transport_class, "__init__") as patched:
            patched.return_value = None
            client = client_class(client_options=options)
            patched.assert_called_once_with(
                credentials=mock_cred,
                credentials_file=None,
                host=client._DEFAULT_ENDPOINT_TEMPLATE.format(
                    UNIVERSE_DOMAIN=client._DEFAULT_UNIVERSE
                ),
                scopes=None,
                client_cert_source_for_mtls=None,
                quota_project_id=None,
                client_info=transports.base.DEFAULT_CLIENT_INFO,
                always_use_jwt_access=True,
                api_audience=None,
            )



# -*- coding: utf-8 -*-
# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from __future__ import annotations

from typing import MutableMapping, MutableSequence

from google.api import monitored_resource_pb2  # type: ignore
import proto  # type: ignore

from google.cloud.monitoring_v3.types import common
from google.cloud.monitoring_v3.types import group as gm_group

__protobuf__ = proto.module(
    package="google.monitoring.v3",
    manifest={
        "ListGroupsRequest",
        "ListGroupsResponse",
        "GetGroupRequest",
        "CreateGroupRequest",
        "UpdateGroupRequest",
        "DeleteGroupRequest",
        "ListGroupMembersRequest",
        "ListGroupMembersResponse",
    },
)


class ListGroupsRequest(proto.Message):
    r"""The ``ListGroup`` request.

    This message has `oneof`_ fields (mutually exclusive fields).
    For each oneof, at most one member field can be set at the same time.
    Setting any member of the oneof automatically clears all other
    members.

    .. _oneof: https://proto-plus-python.readthedocs.io/en/stable/fields.html#oneofs-mutually-exclusive-fields

    Attributes:
        name (str):
            Required. The
            `project <https://cloud.google.com/monitoring/api/v3#project_name>`__
            whose groups are to be listed. The format is:

            ::

                projects/[PROJECT_ID_OR_NUMBER]
        children_of_group (str):
            A group name. The format is:

            ::

                projects/[PROJECT_ID_OR_NUMBER]/groups/[GROUP_ID]

            Returns groups whose ``parent_name`` field contains the
            group name. If no groups have this parent, the results are
            empty.

            This field is a member of `oneof`_ ``filter``.
        ancestors_of_group (str):
            A group name. The format is:

            ::

                projects/[PROJECT_ID_OR_NUMBER]/groups/[GROUP_ID]

            Returns groups that are ancestors of the specified group.
            The groups are returned in order, starting with the
            immediate parent and ending with the most distant ancestor.
            If the specified group has no immediate parent, the results
            are empty.

            This field is a member of `oneof`_ ``filter``.
        descendants_of_group (str):
            A group name. The format is:

            ::

                projects/[PROJECT_ID_OR_NUMBER]/groups/[GROUP_ID]

            Returns the descendants of the specified group. This is a
            superset of the results returned by the
            ``children_of_group`` filter, and includes
            children-of-children, and so forth.

            This field is a member of `oneof`_ ``filter``.
        page_size (int):
            A positive number that is the maximum number
            of results to return.
        page_token (str):
            If this field is not empty then it must contain the
            ``next_page_token`` value returned by a previous call to
            this method. Using this field causes the method to return
            additional results from the previous method call.
    """

    name: str = proto.Field(
        proto.STRING,
        number=7,
    )
    children_of_group: str = proto.Field(
        proto.STRING,
        number=2,
        oneof="filter",
    )
    ancestors_of_group: str = proto.Field(
        proto.STRING,
        number=3,
        oneof="filter",
    )
    descendants_of_group: str = proto.Field(
        proto.STRING,
        number=4,
        oneof="filter",
    )
    page_size: int = proto.Field(
        proto.INT32,
        number=5,
    )
    page_token: str = proto.Field(
        proto.STRING,
        number=6,
    )


class ListGroupsResponse(proto.Message):
    r"""The ``ListGroups`` response.

    Attributes:
        group (MutableSequence[google.cloud.monitoring_v3.types.Group]):
            The groups that match the specified filters.
        next_page_token (str):
            If there are more results than have been returned, then this
            field is set to a non-empty value. To see the additional
            results, use that value as ``page_token`` in the next call
            to this method.
    """

    @property
    def raw_page(self):
        return self

    group: MutableSequence[gm_group.Group] = proto.RepeatedField(
        proto.MESSAGE,
        number=1,
        message=gm_group.Group,
    )
    next_page_token: str = proto.Field(
        proto.STRING,
        number=2,
    )


class GetGroupRequest(proto.Message):
    r"""The ``GetGroup`` request.

    Attributes:
        name (str):
            Required. The group to retrieve. The format is:

            ::

                projects/[PROJECT_ID_OR_NUMBER]/groups/[GROUP_ID]
    """

    name: str = proto.Field(
        proto.STRING,
        number=3,
    )


class CreateGroupRequest(proto.Message):
    r"""The ``CreateGroup`` request.

    Attributes:
        name (str):
            Required. The
            `project <https://cloud.google.com/monitoring/api/v3#project_name>`__
            in which to create the group. The format is:

            ::

                projects/[PROJECT_ID_OR_NUMBER]
        group (google.cloud.monitoring_v3.types.Group):
            Required. A group definition. It is an error to define the
            ``name`` field because the system assigns the name.
        validate_only (bool):
            If true, validate this request but do not
            create the group.
    """

    name: str = proto.Field(
        proto.STRING,
        number=4,
    )
    group: gm_group.Group = proto.Field(
        proto.MESSAGE,
        number=2,
        message=gm_group.Group,
    )
    validate_only: bool = proto.Field(
        proto.BOOL,
        number=3,
    )


class UpdateGroupRequest(proto.Message):
    r"""The ``UpdateGroup`` request.

    Attributes:
        group (google.cloud.monitoring_v3.types.Group):
            Required. The new definition of the group. All fields of the
            existing group, excepting ``name``, are replaced with the
            corresponding fields of this group.
        validate_only (bool):
            If true, validate this request but do not
            update the existing group.
    """

    group: gm_group.Group = proto.Field(
        proto.MESSAGE,
        number=2,
        message=gm_group.Group,
    )
    validate_only: bool = proto.Field(
        proto.BOOL,
        number=3,
    )


class DeleteGroupRequest(proto.Message):
    r"""The ``DeleteGroup`` request. The default behavior is to be able to
    delete a single group without any descendants.

    Attributes:
        name (str):
            Required. The group to delete. The format is:

            ::

                projects/[PROJECT_ID_OR_NUMBER]/groups/[GROUP_ID]
        recursive (bool):
            If this field is true, then the request means
            to delete a group with all its descendants.
            Otherwise, the request means to delete a group
            only when it has no descendants. The default
            value is false.
    """

    name: str = proto.Field(
        proto.STRING,
        number=3,
    )
    recursive: bool = proto.Field(
        proto.BOOL,
        number=4,
    )


class ListGroupMembersRequest(proto.Message):
    r"""The ``ListGroupMembers`` request.

    Attributes:
        name (str):
            Required. The group whose members are listed. The format is:

            ::

                projects/[PROJECT_ID_OR_NUMBER]/groups/[GROUP_ID]
        page_size (int):
            A positive number that is the maximum number
            of results to return.
        page_token (str):
            If this field is not empty then it must contain the
            ``next_page_token`` value returned by a previous call to
            this method. Using this field causes the method to return
            additional results from the previous method call.
        filter (str):
            An optional `list
            filter <https://cloud.google.com/monitoring/api/learn_more#filtering>`__
            describing the members to be returned. The filter may
            reference the type, labels, and metadata of monitored
            resources that comprise the group. For example, to return
            only resources representing Compute Engine VM instances, use
            this filter:

            ::

                `resource.type = "gce_instance"`
        interval (google.cloud.monitoring_v3.types.TimeInterval):
            An optional time interval for which results
            should be returned. Only members that were part
            of the group during the specified interval are
            included in the response.  If no interval is
            provided then the group membership over the last
            minute is returned.
    """

    name: str = proto.Field(
        proto.STRING,
        number=7,
    )
    page_size: int = proto.Field(
        proto.INT32,
        number=3,
    )
    page_token: str = proto.Field(
        proto.STRING,
        number=4,
    )
    filter: str = proto.Field(
        proto.STRING,
        number=5,
    )
    interval: common.TimeInterval = proto.Field(
        proto.MESSAGE,
        number=6,
        message=common.TimeInterval,
    )


class ListGroupMembersResponse(proto.Message):
    r"""The ``ListGroupMembers`` response.

    Attributes:
        members (MutableSequence[google.api.monitored_resource_pb2.MonitoredResource]):
            A set of monitored resources in the group.
        next_page_token (str):
            If there are more results than have been returned, then this
            field is set to a non-empty value. To see the additional
            results, use that value as ``page_token`` in the next call
            to this method.
        total_size (int):
            The total number of elements matching this
            request.
    """

    @property
    def raw_page(self):
        return self

    members: MutableSequence[
        monitored_resource_pb2.MonitoredResource
    ] = proto.RepeatedField(
        proto.MESSAGE,
        number=1,
        message=monitored_resource_pb2.MonitoredResource,
    )
    next_page_token: str = proto.Field(
        proto.STRING,
        number=2,
    )
    total_size: int = proto.Field(
        proto.INT32,
        number=3,
    )


__all__ = tuple(sorted(__protobuf__.manifest))
